# pylint: disable=too-many-arguments
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck

ARL, PFA and CED of Shiryayev-Roberts procedure
'''
import numpy as np
import pandas as pd
from scipy import stats


def shroArlPfaCedNorm(mean0=0, mean1=None, sd=1, n=10, delta=1, tau=None,
                      N=100, limit=10000, seed=None, w=19, verbose=True):
    np.random.seed(seed)

    if delta is None:
        delta = (mean1 - mean0) / sd
    if mean1 is None:
        mean1 = mean0 + delta * sd

    randFunc0 = stats.norm(loc=mean0, scale=sd / np.sqrt(n))
    randFunc1 = stats.norm(loc=mean1, scale=sd / np.sqrt(n))

    rls = []
    for _ in range(N):
        if tau is None:
            x = randFunc0.rvs(limit)
        else:
            x = np.concatenate([randFunc0.rvs(tau), randFunc1.rvs(limit - tau)])
        rls.append(runLengthShroNorm(x, mean0, sd, n, delta, w)['rl'])
    result = {'rls': rls}

    rls = np.ma.masked_invalid(rls)
    result['statistic'] = {
        'ARL': np.mean(rls),
        'Std. Error': np.sqrt((np.mean(rls ** 2) - np.mean(rls) ** 2) / N),
    }
    if tau is not None:
        pfa = np.mean(rls < tau)
        ced = np.mean(rls[rls >= tau]) - tau
        se = np.sqrt(
            (np.sum(rls[rls >= tau] ** 2) / (N - np.sum(rls < tau)) - ced ** 2)
            /
            (N - np.sum(rls < tau)))
        result['statistic']['PFA'] = pfa
        result['statistic']['CED'] = ced
        result['statistic']['CED-Std. Error'] = se

    if verbose:
        print(pd.Series(result['statistic']))
    return result


def runLengthShroNorm(x, mean, sigma, n, delta, ubd):
    limit = len(x)

    wm = 0
    wmv = [None]
    m = 1

    e1 = n * delta / sigma ** 2
    e2 = e1 * delta / 2
    while m < limit and wm < ubd:
        s1 = 0
        wm = 0
        for i in range(0, m):
            s1 = s1 + x[m - i] - mean
            wm = wm + np.exp(s1 * e1 - (i + 1) * e2)
        wmv.append(wm)
        m += 1
    return {'rl': m, 'w': wmv}


def runLengthShroPois(x, rho, delta, ubd):
    limit = len(x)

    wm = 0
    wmv = [None]
    m = 1

    e1 = np.log(rho)
    while m < limit and wm < ubd:
        s1 = 0
        wm = 0
        for i in range(0, m):
            s1 = s1 + x[m - i]
            wm = wm + np.exp(s1 * e1 - (i + 1) * delta)
        wmv.append(wm)
        m += 1
    if m == limit:
        m = np.inf
    return {'rl': m, 'w': wmv}


def shroArlPfaCedPois(lambda0=10, lambda1=None, delta=1, tau=None,
                      N=100, limit=10000, seed=None, w=19, verbose=True):
    np.random.seed(seed)

    if delta is None:
        delta = lambda1 - lambda0
    if lambda1 is None:
        lambda1 = lambda0 + delta

    rho = (lambda0 + delta) / lambda0

    rls = []
    for _ in range(N):
        if tau is None:
            x = stats.poisson(lambda0).rvs(limit)
        else:
            x = np.concatenate([
                stats.poisson(lambda0).rvs(tau),
                stats.poisson(lambda1).rvs(limit - tau),
            ])
        rls.append(runLengthShroPois(x, rho, delta, w)['rl'])
    result = {'rls': rls}

    rls = np.ma.masked_invalid(rls)
    result['statistic'] = {
        'ARL': np.mean(rls),
        'Std. Error': np.sqrt((np.mean(rls ** 2) - np.mean(rls) ** 2) / N),
    }
    if tau is not None:
        pfa = np.mean(rls < tau)
        ced = np.mean(rls[rls >= tau]) - tau
        se = np.sqrt(
            (np.sum(rls[rls >= tau] ** 2) / (N - np.sum(rls < tau)) - ced ** 2)
            /
            (N - np.sum(rls < tau)))
        result['statistic']['PFA'] = pfa
        result['statistic']['CED'] = ced
        result['statistic']['CED-Std. Error'] = se

    if verbose:
        print(pd.Series(result['statistic']))
    return result



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd
from scipy import stats
from scipy.stats import norm

from mistat.simulation.mistatSimulation import SimulationResult

from .mistatSimulation import (MistatSimulation, convert_to_list,
                               repeat_elements)

default_errors = {
    'm': 0.1,
    's': 0.01,
    'v0': 0.00025,
    'k': 50,
    'p0': 0.01,
    't': 0.13,
    't0': 0.13
}


@dataclass
class PistonSimulator(MistatSimulation):  # pylint: disable=too-many-instance-attributes
    m: float = 60
    s: float = 0.02
    v0: float = 0.01
    k: float = 5_000
    p0: float = 110_000
    t: float = 296
    t0: float = 360

    n_simulation: int = 50  # desired number of simulations
    seed: Optional[float] = None
    check: bool = True

    def __post_init__(self):
        if self.seed is not None:
            np.random.seed(seed=self.seed)

        if self.check:
            validate_range(self.m, 30, 60, 'Piston weight m is out of range, [30, 60] kg')
            validate_range(self.s, 0.005, 0.02, "Piston surface area s is out of range, [0.005, 0.020] m^2")
            validate_range(self.v0, 0.002, 0.01, "Value of initial gas volume v0 is out of range, [0.002, 0.010] m^3")
            validate_range(self.k, 1000, 5000, "Value of spring coefficient k is out of range, [1000, 5000] N/m")
            validate_range(self.p0, 90_000, 110_000,
                           "Value of atmospheric pressure p0 is out of range, [90000, 110000] N/m^2")
            validate_range(self.t, 290, 296, "Value of ambient temperature t is out of range, [290, 296] K")
            validate_range(self.t0, 340, 360, "Value of filling gas temperature t0 is out of range, [340, 360] K")

        if self.n_simulation < 1:
            raise ValueError('Number of simulations must be greater 1')

        # Convert to lists
        maxsize = 0
        for option in ('m', 's', 'v0', 'k', 'p0', 't', 't0'):
            values = convert_to_list(getattr(self, option))
            maxsize = max(maxsize, len(values))
            setattr(self, option, values)
        if maxsize == 1:
            maxsize = self.n_simulation

        # Make sure that the vectors are all the same length
        for option in ('m', 's', 'v0', 'k', 'p0', 't', 't0'):
            values = getattr(self, option)
            if maxsize % len(values) != 0:
                raise ValueError(f'Inconsistent length of option {option}')
            values = np.array(list(repeat_elements(values, maxsize // len(values))))
            setattr(self, option, values)

        self.errors = dict(default_errors)

    def simulate(self):
        size = len(self.m)

        # add errors
        errors = self.errors
        m_s = self.m + norm.rvs(size=size, loc=0, scale=errors['m'])
        s_s = self.s + norm.rvs(size=size, loc=0, scale=errors['s'])
        s_s[s_s < 0.00001] = 0.00001
        v0_s = self.v0 + norm.rvs(size=size, loc=0, scale=errors['v0'])
        v0_s[v0_s < 0] = 0.001
        k_s = self.k + norm.rvs(size=size, loc=0, scale=errors['k'])
        p0_s = self.p0 + norm.rvs(size=size, loc=0, scale=errors['p0'])
        p0_s[p0_s < 0] = 0.001
        t_s = self.t + norm.rvs(size=size, loc=0, scale=errors['t'])
        t0_s = self.t0 + norm.rvs(size=size, loc=0, scale=errors['t0'])

        Mg = 0.2 * k_s
        A = p0_s * s_s + 2 * Mg - k_s * v0_s / s_s
        V = s_s * (np.sqrt(A**2 + 4 * k_s * p0_s * v0_s * t_s / t0_s) - A) / (2 * k_s)
        res = 2 * np.pi * np.sqrt(m_s / (k_s + ((s_s**2) * p0_s * v0_s * t_s) / (t0_s * V * V)))

        result = {option: getattr(self, option) for option in ('m', 's', 'v0', 'k', 'p0', 't', 't0')}
        result['seconds'] = res
        return SimulationResult(result)

    @staticmethod
    def cycleTime(m, s, v0, k, p0, t, t0):  # pylint: disable=too-many-arguments
        Mg = 0.2 * k
        A = p0 * s + 2 * Mg - k * v0 / s
        V = s * (np.sqrt(A**2 + 4 * k * p0 * v0 * t / t0) - A) / (2 * k)
        return 2 * np.pi * np.sqrt(m / (k + ((s**2) * p0 * v0 * t) / (t0 * V * V)))


def validate_range(value, left, right, message):
    if isinstance(value, (tuple, list, pd.core.series.Series, np.ndarray)):
        left = left - 1e-6
        right = right + 1e-6
        if not all(left <= v <= right for v in value):
            raise ValueError(message)
    else:
        if not left <= value <= right:
            raise ValueError(message)


def uniformSumDistribution(size=1, k=6, left=0, right=1):
    """ return size random numbers from the uniform sum distribution [left, right] based on k sums """
    if right <= left or size < 1 or k < 1:
        raise ValueError('Invalid arguments')
    factor = (right - left) / k
    return np.array([np.sum(stats.uniform.rvs(size=k)) * factor + left for _ in range(size)])



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numbers
from collections import namedtuple
from enum import Enum

import numpy as np
import pandas as pd
from scipy import stats
from scipy.special import gammaln  # pylint: disable=no-name-in-module

GroupMeans = namedtuple('GroupMeans', 'statistics,center')


class QCCStatistics:

    def __init__(self):
        self.default = None
        self.statistics = {}
        self.register(Xbar_statistic, default=True)
        self.register(Xbar_one_statistic)
        self.register(R_statistic)
        self.register(S_statistic)
        self.register(P_statistic)
        self.register(NP_statistic)

    def get(self, name):
        if name is None and self.default is None:
            raise ValueError('you need give a name as default is not defined')
        if name.lower() not in self.statistics:
            raise ValueError(f'unknown statistic {name}')
        return self.statistics[name.lower()]()

    def register(self, statistic, default=False):
        name = statistic.qcc_type
        if name.lower() in self.statistics:
            raise ValueError(f'duplicate definition of statistics {name}')
        if default and self.default is not None:
            raise ValueError(f'duplicate definition of default statistics {name} and {default}')
        name = name.lower()
        if default:
            self.default = name
        self.statistics[name] = statistic

    def __iter__(self):
        if self.default is not None:
            yield self.statistics[self.default]
        for statistic in self.statistics.values():
            if statistic.qcc_type.lower() == self.default:
                continue
            yield statistic


class SD_estimator(Enum):
    # UnWeighted AVErage of subgroup estimates based on subgroup Ranges
    uwave_r = 'UWAVE-R'
    # UnWeighted AVErage of subgroup estimates based on subgroup Standard Deviations
    uwave_sd = 'UWAVE-SD'
    # Minimum Variance Linear Unbiased Estimator computed as a weighted average of subgroups
    # estimates based on subgroup Ranges
    mvlue_r = 'MVLUE-R'
    # Minimum Variance Linear Unbiased Estimator computed as a weighted average of subgroup
    # estimates based on subgroup Standard Deviations
    mvlue_sd = 'MVLUE-SD'
    # Root-Mean-Square estimator computed as a weighted average of subgroup estimates based
    # on subgroup Standard Deviations
    rmsdf = 'RMSDF'
    mr = 'MR'
    sd = 'SD'

    @classmethod
    def get(cls, name, default):
        if name is None:
            name = default
        if isinstance(name, cls):
            return name
        if not isinstance(name, str):
            raise ValueError(f'name of estimator must be string or SD_estimator, found: {name}')
        name = name.upper()
        for estimator in cls:
            if estimator.value == name:
                return estimator
        raise ValueError(f'unknown SD estimator {name}')


# exp.R.unscaled a vector specifying, for each sample size, the expected value of the relative range
# (i.e. R/σ) for a normal distribution. This appears as d2 on most tables containing factors for
# the construction of control charts.
_exp_R_unscaled = [np.NaN, np.NaN, 1.128, 1.693, 2.059, 2.326, 2.534, 2.704, 2.847, 2.970, 3.078, 3.173,
                   3.258, 3.336, 3.407, 3.472, 3.532, 3.588, 3.640, 3.689, 3.735, 3.778, 3.819, 3.858, 3.895, 3.931]
# se.R.unscaled a vector specifying, for each sample size, the standard error of the relative range
# (i.e. R/σ) for a normal distribution. This appears as d3 on most tables containing factors for
# the construction of control charts.
_se_R_unscaled = [np.NaN, np.NaN, 0.8525033, 0.8883697, 0.8798108, 0.8640855, 0.8480442, 0.8332108, 0.8198378,
                  0.8078413, 0.7970584, 0.7873230, 0.7784873, 0.7704257, 0.7630330, 0.7562217, 0.7499188, 0.7440627,
                  0.7386021, 0.7334929, 0.7286980, 0.7241851, 0.7199267, 0.7158987, 0.7120802, 0.7084528, 0.7050004,
                  0.7017086, 0.6985648, 0.6955576, 0.6926770, 0.6899137, 0.6872596, 0.6847074, 0.6822502, 0.6798821,
                  0.6775973, 0.6753910, 0.6732584, 0.6711952, 0.6691976, 0.6672619, 0.6653848, 0.6635632, 0.6617943,
                  0.6600754, 0.6584041, 0.6567780, 0.6551950, 0.6536532, 0.6521506]


class Base_statistic:
    @staticmethod
    def getSizes(data):
        return [int(n) for n in (~np.isnan(data)).sum(1)]

    def stats(self, data, sizes=None):
        raise NotImplementedError()

    def sd(self, data, std_dev, sizes=None):
        raise NotImplementedError()

    def limits(self, center, std_dev, sizes, conf):
        raise NotImplementedError()


class Xbar_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart xbar chart """
    qcc_type = 'xbar'
    description = ('mean', 'means of a continuous process variable')

    def stats(self, data, sizes=None):
        return GroupMeans(np.nanmean(data, axis=1), np.nanmean(data))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if sizes is None:
            sizes = self.getSizes(data)

        # set default method if std_dev is None
        if std_dev is None:
            std_dev = SD_estimator.uwave_r
            if max(sizes) > 25:
                std_dev = SD_estimator.rmsdf

        if std_dev == SD_estimator.uwave_r:
            r = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
            d2 = [_exp_R_unscaled[size] for size in sizes]
            return np.mean(r / d2)
        if std_dev == SD_estimator.uwave_sd:
            s = np.nanstd(data, axis=1, ddof=1)
            c4 = np.array([qcc_c4(size) for size in sizes])
            return np.mean(s / c4)
        if std_dev == SD_estimator.mvlue_r:
            r = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
            d2 = np.array([_exp_R_unscaled[size] for size in sizes])
            d3 = np.array([_se_R_unscaled[size] for size in sizes])
            w = (d2 / d3) ** 2
            return np.sum(w * r / d2) / np.sum(w)
        if std_dev == SD_estimator.mvlue_sd:
            s = np.nanstd(data, axis=1, ddof=1)
            c4 = np.array([qcc_c4(size) for size in sizes])
            w = c4 ** 2 / (1 - c4**2)
            return np.sum(w * s / c4) / np.sum(w)
        if std_dev == SD_estimator.rmsdf:
            s = np.nanstd(data, axis=1, ddof=1)
            w = np.array([size - 1 for size in sizes])
            return np.sqrt(np.sum(w * s**2) / np.sum(w)) / qcc_c4(np.sum(w) + 1)

        raise NotImplementedError()

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]

        se_stats = std_dev / np.sqrt(sizes)
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            sigmas = stats.norm.ppf(1 - (1 - conf) / 2)
            lcl = center - sigmas * se_stats
            ucl = center + sigmas * se_stats
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class Xbar_one_statistic(Base_statistic):
    """ Statistics used in xbar-one charts """
    qcc_type = 'xbarone'
    description = ('mean', 'one-at-time data of a continuous process variable')

    @staticmethod
    def getSizes(data):
        return [1 for _ in data]

    def stats(self, data, sizes=None):
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        return GroupMeans(np.array(data).flatten(), np.mean(data))

    def sd(self, data, std_dev=None, sizes=None, k=2):  # pylint: disable=unused-import, arguments-differ
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values

        # set default method if std_dev is None
        std_dev = SD_estimator.get(std_dev, SD_estimator.mr)
        std_dev = std_dev or SD_estimator.mr
        if isinstance(std_dev, str):
            std_dev = std_dev.upper()

        if std_dev == SD_estimator.mr:
            d = 0
            for i in range(len(data) - (k - 1)):
                group = data[i:i + k]
                d += np.max(group) - np.min(group)
            return d / ((len(data) - k + 1) * _exp_R_unscaled[k])

        if std_dev == SD_estimator.sd:
            return np.std(data, ddof=1) / qcc_c4(len(data))

        raise NotImplementedError(f'estimator {std_dev}')

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')
        se_stats = std_dev
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            sigmas = stats.norm.ppf(1 - (1 - conf) / 2)
            lcl = center - sigmas * se_stats
            ucl = center + sigmas * se_stats
        if isinstance(lcl, numbers.Number):
            lcl = [lcl]
            ucl = [ucl]
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class R_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart R chart """
    qcc_type = 'R'
    description = ('range', ' ranges of a continuous process variable')

    def stats(self, data, sizes=None):
        if sizes is None:
            sizes = self.getSizes(data)
        by_group = np.nanmax(data, axis=1) - np.nanmin(data, axis=1)
        return GroupMeans(by_group, sum(sizes * by_group) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if std_dev is None:
            std_dev = SD_estimator.uwave_r

        if std_dev not in (SD_estimator.uwave_r, SD_estimator.mvlue_r):
            raise ValueError(f'invalid std_dev method {std_dev}')
        return Xbar_statistic().sd(data, std_dev=std_dev, sizes=sizes)

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        Rtab = len(_se_R_unscaled)

        if conf >= 1:
            if any(size > Rtab for size in sizes):
                raise ValueError(f'group size must be less than {Rtab + 1} when giving nsigmas')
            seR = std_dev * np.array([_se_R_unscaled[size] for size in sizes])
            lcl = center - conf * seR
            ucl = center + conf * seR
        else:
            # ucl <- qtukey(1 - (1 - conf)/2, sizes, 1e100) * std.dev
            # lcl <- qtukey((1 - conf)/2, sizes, 1e100) * std.dev
            raise NotImplementedError(
                'Tukey studentized range distribution not available from scipy. Use nsigmas instead')
        lcl[lcl < 0] = 0
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class S_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart S chart """
    qcc_type = 'S'
    description = ('standard deviation', 'standard deviations of a continuous variable')

    def stats(self, data, sizes=None):
        if sizes is None:
            sizes = self.getSizes(data)
        by_group = np.nanstd(data, ddof=1, axis=1)
        return GroupMeans(by_group, sum(sizes * by_group) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev

        if std_dev is None:
            std_dev = SD_estimator.uwave_sd

        if std_dev not in (SD_estimator.uwave_sd, SD_estimator.mvlue_sd, SD_estimator.rmsdf):
            raise ValueError(f'invalid std_dev method {std_dev}')
        return Xbar_statistic().sd(data, std_dev=std_dev, sizes=sizes)

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        c4 = np.array([qcc_c4(size) for size in sizes])
        se_stats = std_dev * np.sqrt(1 - c4**2)
        if conf >= 1:
            lcl = center - conf * se_stats
            ucl = center + conf * se_stats
        else:
            p = (1 - conf) / 2
            chi2 = [stats.chi2(size - 1).ppf(p) for size in sizes]
            lcl = std_dev * np.sqrt(chi2 / (sizes - 1))

            chi2 = [stats.chi2(size - 1).ppf(1 - p) for size in sizes]
            ucl = std_dev * np.sqrt(chi2 / (sizes - 1))
        lcl[lcl < 0] = 0
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})


class P_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart p chart """
    qcc_type = 'p'
    description = ('proportion', 'proportion of nonconforming units')

    @staticmethod
    def getSizes(data):
        raise ValueError('p-charts require argument sizes to be provided')

    def stats(self, data, sizes=None):
        if sizes is None:
            raise ValueError('p-charts require argument sizes to be provided')
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        elif isinstance(sizes, (pd.Series, pd.DataFrame)):
            sizes = sizes.values
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        data = data.flatten()
        sizes = sizes.flatten()
        return GroupMeans(data / sizes, sum(data) / sum(sizes))

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        else:
            sizes = np.array(sizes)
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values
        pbar = np.sum(data) / np.sum(sizes)
        return np.sqrt(pbar * (1 - pbar))

    def limits(self, center, std_dev, sizes, conf):
        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)
        limits = NP_statistic().limits(center * sizes, std_dev, sizes, conf)

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        limits['LCL'] = limits['LCL'] / sizes
        limits['UCL'] = limits['UCL'] / sizes
        return limits


class NP_statistic(Base_statistic):
    """ Statistics used in computing and drawing a Shewhart np chart """
    qcc_type = 'np'
    description = ('count', 'number of nonconforming units')

    def stats(self, data, sizes=None):
        if sizes is None:
            raise ValueError('np-charts require argument sizes to be provided')
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        pbar = np.sum(data) / np.sum(sizes)
        center = sizes * pbar
        if all(c == center[0] for c in center):
            center = center[0]
        return GroupMeans(data, center)

    def sd(self, data, std_dev=None, sizes=None):
        if isinstance(std_dev, numbers.Number):
            return std_dev
        if isinstance(sizes, numbers.Number):
            sizes = np.array([sizes] * len(data))
        else:
            sizes = np.array(sizes)
        if isinstance(data, (pd.Series, pd.DataFrame)):
            data = data.values

        pbar = np.sum(data) / np.sum(sizes)
        std_dev = np.sqrt(sizes * pbar * (1 - pbar))
        if all(c == std_dev[0] for c in std_dev):
            std_dev = std_dev[0]
        return std_dev

    def limits(self, center, std_dev, sizes, conf):
        if conf < 0:
            raise ValueError(f'invalid conf argument {conf}')

        if all(size == sizes[0] for size in sizes):
            sizes = [sizes[0]]
        sizes = np.array(sizes)

        pbar = np.mean(center / sizes)
        if conf >= 1:
            tol = conf * np.sqrt(pbar * (1 - pbar) * sizes)
            lcl = center - tol
            ucl = center + tol
        else:
            p = (1 - conf) / 2
            lcl = np.array([stats.binom(size, p).ppf(pbar) for size in sizes])
            ucl = np.array([stats.binom(size, p).isf(pbar) for size in sizes])
        lcl[lcl < 0] = 0
        ucl[ucl > sizes] = sizes
        return pd.DataFrame({'LCL': lcl, 'UCL': ucl})

# # c Chart
#
# stats.c <- function(data, sizes)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   if (length(unique(sizes)) != 1)
#      stop("all sizes must be be equal for a c chart")
#   statistics <- data
#   center <- mean(statistics)
#   list(statistics = statistics, center = center)
# }
#
# sd.c <- function(data, sizes, ...)
# {
#   data <- as.vector(data)
#   std.dev <- sqrt(mean(data))
#   return(std.dev)
# }
#
# limits.c <- function(center, std.dev, sizes, conf)
# {
#   if (conf >= 1)
#      { lcl <- center - conf * sqrt(center)
#        lcl[lcl < 0] <- 0
#        ucl <- center + conf * sqrt(center)
#      }
#   else
#      { if (conf > 0 & conf < 1)
#           { ucl <- qpois(1 - (1 - conf)/2, center)
#             lcl <- qpois((1 - conf)/2, center)
#           }
#        else stop("invalid conf argument. See help.")
#      }
#   limits <- matrix(c(lcl, ucl), ncol = 2)
#   rownames(limits) <- rep("", length = nrow(limits))
#   colnames(limits) <- c("LCL", "UCL")
#   return(limits)
# }
#
# # u Chart
#
# stats.u <- function(data, sizes)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   statistics <- data/sizes
#   center <- sum(sizes * statistics)/sum(sizes)
#   list(statistics = statistics, center = center)
# }
#
# sd.u <- function(data, sizes, ...)
# {
#   data <- as.vector(data)
#   sizes <- as.vector(sizes)
#   std.dev <- sqrt(sum(data)/sum(sizes))
#   return(std.dev)
# }
#
# limits.u <- function(center, std.dev, sizes, conf)
# {
#   sizes <- as.vector(sizes)
#   if (length(unique(sizes))==1)
#      sizes <- sizes[1]
#   limits.c(center * sizes, std.dev, sizes, conf) / sizes
# }


def qcc_c4(n):
    return np.sqrt(2 / (n - 1)) * np.exp(gammaln(n / 2) - gammaln((n - 1) / 2))


qccStatistics = QCCStatistics()



# mypy: disallow_untyped_defs,disallow_untyped_calls
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from typing import List, Union

import numpy as np

from .dodge_base import AcceptanceSamplingPlan

# Chain Sampling Plans for the binomial and Poisson distributions.
# #'
# #'
# #' @aliases ChainSamplingPlans ChainBinomial ChainPoisson
# #' @param N the lot size
# #' @param n the sample size
# #' @param i the number of preceding lots that are free from nonconforming units
# #' for the lot to be accepted
# #' @param p a vector of values for the possible fraction of product that is
# #' nonconforming
# #' @param Plots logical to request generation of the four plots
# #' @return A matrix containing the argument \code{p} as supplied and the
# #' calculated OC, ATI and ???
# #' @author Raj Govindaraju with minor editing by Jonathan Godfrey
# #' @references Dodge, H.F. (1955) \dQuote{Chain Sampling Inspection Plan},
# #' \emph{Industrial Quality Control} \bold{11}(4), pp10-13.


def ChainPlanBinomial(N: int, n: int, i: int, p: Union[List[float], np.ndarray] = None) -> AcceptanceSamplingPlan:
    if p is None:
        p = np.arange(0, 0.301, 0.001)
    p = np.array(p)

    OC = (1 - p)**n + n * p * (1 - p)**(n + n * i - 1)
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=list(p), OC=list(OC), n=[n] * len(p), AOQ=list(AOQ), ATI=list(ATI), ASN=None)


def ChainPlanPoisson(N: int, n: int, i: int, p: Union[List[float], np.ndarray] = None) -> AcceptanceSamplingPlan:
    p = p if p is not None else np.arange(0, 0.3, 0.001)
    p = np.array(p)

    OC = np.exp(-n * p) + n * p * np.exp(-n * p * (i + 1))
    AOQ = (N - n) * p * OC / N
    ATI = n * OC + N * (1 - OC)
    return AcceptanceSamplingPlan(p=list(p), OC=list(OC), n=[n] * len(p), AOQ=list(AOQ), ATI=list(ATI), ASN=None)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from collections import namedtuple

import numpy as np
from scipy import stats
from scipy.linalg import inv, toeplitz
from statsmodels.tsa.stattools import acf


def optimalLinearPredictor(x, n, nlags=10):
    ns = len(x)
    res = np.array(x)
    for i in range(n + 1, ns + 1):
        xs = x[:i]
        ak = acf(xs, nlags=nlags, fft=True)
        R = toeplitz(ak[:nlags])
        r = ak[1:]
        b = np.matmul(inv(R), r)
        res[i - 1] = np.sum(x[(i - 2):(i - nlags - 2):-1] * b)
    return res


def quadraticPredictor(x, n, s):
    ns = len(x)
    res = np.array(x)
    nx = np.arange(0, -n, -1)
    A = np.column_stack([[1] * n, nx, nx**2])
    invAA = inv(np.matmul(A.transpose(), A))
    invAAA = np.matmul(invAA, A.transpose())
    ts = np.array([1, s, s**2])
    for i in range(n, ns - s + 1):
        xs = x[i - n:i]
        xs = xs[::-1]
        b = np.matmul(invAAA, xs)
        pred = np.sum(b * ts)
        res[i + s - 1] = pred
    return res


def masPredictor(x, m, s):
    n = len(x)
    wm = np.arange(-m, m + 1)
    res = np.array(x)
    for i in range(m, n - m - s):
        xtm = x[i - m:i + m + 1]
        b0m = np.mean(xtm)
        b1m = 3 * np.sum(xtm * wm) / (m * (m + 1) * (2 * m + 1))
        res[i + m + s] = b0m + (m + s) * b1m
    return res


def normRandomWalk(n, v, w, c, seed=None):
    np.random.seed(seed)
    s1 = np.sqrt(v)
    s2 = np.sqrt(w)
    X = np.zeros(n)
    Tet = np.zeros(n)
    cf = np.zeros(n)
    Mt = np.zeros(n)
    cf[0] = (c + w) * v / (c + w + v)
    Tet[0] = s2 * stats.norm.rvs()
    X[0] = Tet[0] + s1 * stats.norm.rvs()
    for i in range(1, n):
        Tet[i] = Tet[i - 1] + s2 * stats.norm.rvs()
        X[i] = Tet[i] + s1 * stats.norm.rvs()
        cf[i] = (cf[i - 1] + w) * v / (cf[i - 1] + w + v)
        at = (cf[i] + w) / (cf[i] + w + v)
        Mt[i] = (1 - at) * Mt[i - 1] + at * X[i]
    nvrmResult = namedtuple('nvrmResult', 't,X,predicted')
    return nvrmResult(np.arange(n), X, Mt)


def dlmLinearGrowth(X, C0, v, W, M0):
    n = len(X)
    pred = np.ones(n)
    for i in range(n):
        at = np.array([1.0, i + 1]).transpose()
        pred[i] = np.matmul(at.transpose(), M0)
        ut = v + np.matmul(at.transpose(), np.matmul(C0 + W, at))
        ei = X[i] - np.matmul(at.transpose(), M0)
        ev = np.array([ei / ut, (i + 1) * ei / ut]).transpose()
        M0 = M0 + np.matmul((C0 + W), ev)
        B = np.array([[1 / ut, (i + 1) / ut], [(i + 1) / ut, (i + 1)**2 / ut]])
        C0 = (C0 + W) - np.matmul(C0 + W, np.matmul(B, (C0 + W)))
    return pred


def simulateARMA(n, a, b, seed=None):
    np.random.seed(seed)
    p = len(a)
    q = len(b)
    pm = max(p, q)
    ns = n + pm

    X = np.zeros(ns)
    X[0] = stats.norm.rvs()
    for i in range(1, pm):
        X[i] = X[i - 1] + stats.norm.rvs()
    E = stats.norm.rvs(size=ns)

    for i in range(pm, ns):
        X[i] = np.sum(X[i - p:i][::-1] * a)
        E[i] = np.sum(E[i - q:i][::-1] * b) + E[i]
    X = X + E
    return X


def predictARMA(X, a):
    n = len(X)
    p = len(a)
    pred = np.zeros(n)
    pred[:p] = X[:p]
    for i in range(p, n):
        pred[i] = np.sum(X[i - p:i][::-1] * a)
    return pred



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import itertools
from itertools import groupby
from numbers import Number

import numpy as np
import pandas as pd

RUN_LENGTH = 7


# Functions used to signal points out of control

def shewhartRules(qcc, **kwargs):
    return {
        **beyondLimits(qcc, **kwargs),
        **violatingRuns(qcc, **kwargs),
    }


def beyondLimits(qcc, limits=None, **kwargs):
    limits = qcc.limits if limits is None else limits
    statistics = list(qcc.stats.statistics)
    if qcc.newstats:
        statistics.extend(qcc.newstats.statistics)
    if isinstance(statistics, pd.Series):
        statistics = statistics.values
    statistics = np.array(statistics).flatten()

    if len(limits['UCL']) == 1:
        return {'beyondLimits': {'UCL': np.nonzero(statistics > limits['UCL'][0])[0],
                                 'LCL': np.nonzero(statistics < limits['LCL'][0])[0]}}
    return {'beyondLimits': {'UCL': np.nonzero(statistics.ravel() > limits['UCL'].ravel())[0],
                             'LCL': np.nonzero(statistics.ravel() < limits['LCL'].ravel())[0]}}


def violatingRuns(qcc, run_length=7, **kwargs):
    if run_length == 0:
        return {'violatingRuns': []}
    center = qcc.center
    statistics = list(qcc.stats.statistics)
    if qcc.newstats:
        statistics.extend(qcc.newstats.statistics)
    if isinstance(statistics, pd.Series):
        statistics = statistics.values
    statistics = np.array(statistics).flatten()
    if isinstance(center, Number) or len(center.shape) == 0:
        diffs = statistics - center
    elif center.shape[0] == 1:
        diffs = statistics - center[0]
    else:
        raise NotImplementedError()
    diffs[diffs > 0] = 1
    diffs[diffs < 0] = -1
    runs = run_length_encoding(diffs, as_list=True)
    violating = runs >= run_length
    violating_above = violating & (diffs > 0)
    violating_below = violating & (diffs < 0)

    violators = []
    for violating in (violating_above, violating_below):
        current = 0
        for length, violates in run_length_encoding(violating):
            if violates:
                violators.extend(range(current + run_length - 1, current + length))
            current += length
    return {'violatingRuns': violators}


def run_length_encoding(sequence, as_list=False):
    rle = [(len(list(g)), k) for k, g in groupby(sequence)]
    if as_list:
        def flatten(x):
            return itertools.chain(*x)
        rle = np.array(list(flatten([count] * count for count, _ in rle)))
    return rle



# pylint: disable=too-many-arguments,too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from collections import namedtuple
from functools import lru_cache

import numpy as np
from scipy import stats

# N = 10
# p = 0.95
# al = 0.5
# k = 10
# gam = 0.95
# Ns = 1000

Stats = namedtuple('Stats', 'mean,std')
SimulateOABResult = namedtuple('SimulateOABResult', 'mgamma,reward')
OptimalOABResult = namedtuple('OptimalOABResult', 'max_reward,rewards')

# Use a cache to save results of call to the beta.cdf function
# to speed up calculations


@lru_cache(maxsize=500)
def cachedBetaCDF(al, a, b):
    return stats.beta.cdf(al, a, b)


def simulateOAB(N, p, al, k, gam, Ns, seed=None):
    if seed is not None:
        np.random.seed(seed)
    res = []
    for _ in range(Ns):
        n = k
        X = stats.binom.rvs(k, p)
        for n in range(k, N + 1):
            prob = cachedBetaCDF(al, X + 1, n + 1 - X)
            if prob > gam:
                break
            X = X + stats.binom.rvs(1, p)
        res.append([n, X + (N - n) * al])
    res = np.array(res)
    means = np.mean(res, axis=0)
    std = np.std(res, axis=0)
    return SimulateOABResult(Stats(means[0], std[0]), Stats(means[1], std[1]))


def optimalOAB(N, al):
    '''
    N= Number of trials; al= Known probability of success in arm A.
    Output: Table of maximal predicted rewards
    '''
    reward = np.zeros([N, N + 1])
    for X in range(N + 1):
        if X < al * (N + 1) - 1:
            reward[0, X] = al
        else:
            reward[0, X] = (X + 1) / (N + 1)
    for j in range(1, N):
        n = N + 1 - (j + 1)
        for i in range(1, n + 2):
            X = i - 1
            cr1 = (X + 1) * reward[j - 1, X + 1] / (N - j + 1)
            cr2 = (N - j - X) * reward[j - 1, X] / (N - j + 1)
            cr3 = (X + 1) / (N - j + 1)
            cr = cr1 + cr2 + cr3
            if cr < al * (j + 1):
                reward[j, i - 1] = al * (j + 1)
            else:
                reward[j, i - 1] = cr

    return OptimalOABResult((reward[N - 1, 0] + reward[N - 1, 1]) / 2, reward)


def optimalOAB2(N, lambda_):
    ''' vectorized version of optimalOAB2 '''
    reward = np.zeros([N, N + 1])
    maxreward = lambda_
    X = np.arange(0, N + 1)
    rho = (X + 1) / (N + 1)
    rho = np.maximum(maxreward, rho)
    reward[0, :] = rho
    for n in range(1, N):
        # calculate the components of rho
        maxreward = lambda_ * (n + 1)
        cr1 = (X[:-1] + 1) / (N - n + 1) * rho[1:]
        cr2 = (N - n - X[:-1]) / (N - n + 1) * rho[:-1]
        cr3 = (X[:-1] + 1) / (N - n + 1)
        # calculate the new rho
        rho = np.maximum(maxreward, cr1 + cr2 + cr3)
        reward[n, :len(rho)] = rho
        # for next iteration, shorten vector X
        X = X[:-1]
    return OptimalOABResult((reward[N - 1, 0] + reward[N - 1, 1]) / 2, reward)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

# R: dodge.CurtBinomial


def curtailedBinomial(n, Ac, p=None):
    """ The ASN Function for Curtailed Single Sampling by Attributes
    Anders Hald and Uffe Msller
    Technometrics, Vol. 18, No. 3(Aug., 1976), pp. 307-312
    """
    p = p if p is not None else np.arange(0, 0.5, 0.01)
    p = np.array(p)

    # handle the edge cases of p = 0, 1
    mask0 = np.where(p == 0)
    mask1 = np.where(p == 1)
    p[mask0] = 1e-6
    p[mask1] = 0.999999999
    q = 1-p
    t1 = stats.binom.cdf(Ac, n+1, p) * (n - Ac) / (n * q)
    t2 = (1 - stats.binom.cdf(Ac+1, n+1, p)) * (Ac + 1) / (n * p)
    t1[mask0] = (n - Ac) / n
    t1[mask1] = 0
    t2[mask0] = 0
    t2[mask1] = (Ac + 1) / n
    ASNfull = n * (t1 + t2)

    p[mask0] = 0
    t1 = stats.binom.cdf(Ac, n, p)
    ASNsemi = n * (t1 + t2)

    return CurtailedSamplePlan(ASNsemi=ASNsemi, ASNfull=ASNfull, p=p, n=n)


@dataclass
class CurtailedSamplePlan:
    p: List[float]
    ASNsemi: List[float]
    ASNfull: List[float]
    n: int

    oc_type: str = 'binomial'

    def __repr__(self):
        return str(pd.DataFrame({
            'p': self.p,
            'ASNsemi': self.ASNsemi,
            'ASNfull': self.ASNfull,
        }))

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots()
        ax.plot(self.p, self.ASNfull, linestyle='--', color='C1', label='Fully curtailed ASN')
        ax.plot(self.p, self.ASNsemi, color='C0', label='Semi-curtailed ASN')

        ax.set_ylim(1, 1.1 * self.n)
        ax.set_xlabel(r'$p$')
        ax.set_ylabel('ASN')
        return ax



# pylint: disable=too-many-arguments, too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from mistat.qcc.qualityControlChart import QualityControlChart


class ProcessCapability:
    """ Computes process capability indices for a qcc object of type 'xbar' """

    def __init__(self, qcc, spec_limits, std_dev=None, target=None, nsigmas=None,
                 confidence_level=0.95):
        if not isinstance(qcc, QualityControlChart):
            raise ValueError('Argument qcc must be a QualityControlChart object')
        if qcc.qcc_type not in ('xbar', 'xbarone'):
            raise ValueError('Process Capability Analysis only available for charts type "xbar" and "xbar.one" charts')

        self.qcc = qcc
        data = self.qcc.data
        self.nobs = np.count_nonzero(~np.isnan(data))
        self.center = qcc.center
        self.std_dev = std_dev or qcc.std_dev

        lsl = spec_limits[0]
        usl = spec_limits[1]
        self.spec_limits = pd.DataFrame({'LSL': [lsl], 'USL': [usl]})

        self.has_target = target is not None
        self.target = target or np.nanmean(self.spec_limits)
        if target is None:
            self.target = np.nanmean(self.spec_limits)
            self.has_target = True

        # validation:
        # - only one of lcl or ucl can be NaN
        # - target must be between self.spec_limits

        self.nsigmas = nsigmas or self.qcc.nsigmas
        if not self.nsigmas:
            raise ValueError('nsigmas not available from qcc object. Provide nsigmas as argument.')

        self.confidence_level = confidence_level
        self.calcPCIndices()

        self.exp_LSL = np.NaN
        if not np.isnan(lsl):
            self.exp_LSL = stats.norm.cdf((lsl - self.center) / self.std_dev) * 100
        self.exp_USL = np.NaN
        if not np.isnan(usl):
            self.exp_USL = (1 - stats.norm.cdf((usl - self.center) / self.std_dev)) * 100
        self.obs_LSL = np.nanmean(data < lsl) * 100
        self.obs_USL = np.nanmean(data > usl) * 100

    def calcPCIndices(self):
        usl = self.spec_limits.values[0, 1]
        lsl = self.spec_limits.values[0, 0]
        nsigmas = self.nsigmas
        std_dev = self.std_dev
        center = self.center
        target = self.target

        self.Cp = (usl - lsl) / (2 * nsigmas * std_dev)
        self.Cp_u = (usl - self.center) / (nsigmas * std_dev)
        self.Cp_l = (self.center - lsl) / (nsigmas * std_dev)
        self.Cp_k = min(self.Cp_u, self.Cp_l)
        self.Cpm = self.Cp / np.sqrt(1 + ((center - target) / std_dev) ** 2)

        alpha = 1 - self.confidence_level
        alphas = [alpha / 2, 1 - alpha / 2]
        n = self.nobs
        self.Cp_limits = [self.Cp * np.sqrt(stats.chi2(n - 1).ppf(a) / (n - 1)) for a in alphas]
        self.Cp_l_limits = [self.Cp_l * (1 + f * stats.norm.ppf(self.confidence_level) *
                                         np.sqrt(1 / (9 * n * self.Cp_l**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]
        self.Cp_u_limits = [self.Cp_u * (1 + f * stats.norm.ppf(self.confidence_level) *
                                         np.sqrt(1 / (9 * n * self.Cp_u**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]
        self.Cp_k_limits = [self.Cp_k * (1 + f * stats.norm.ppf(1 - alpha / 2) *
                                         np.sqrt(1 / (9 * n * self.Cp_k**2) + 1 / (2 * (n - 1))))
                            for f in (-1, 1)]

        ratio = (center - target) / std_dev
        df = n * (1 + ratio**2) / (1 + 2 * ratio**2)
        self.Cpm_limits = [self.Cpm * np.sqrt(stats.chi2(df).ppf(a) / df) for a in alphas]

    def summary(self):
        print('Process Capability Analysis')
        print()
        print(f'Number of obs = {self.nobs:<4d}         Target = {self.target:.2f}')
        print(f'       Center = {self.center:<14.2f}  LSL = {self.spec_limits.values[0, 0]:.2f}')
        print(f'       StdDev = {self.std_dev:<14.6f}  USL = {self.spec_limits.values[0, 1]:.2f}')
        print()
        print('Capability indices:')
        print()
        print('        Value     2.5%   97.5%')
        for key in ('Cp', 'Cp_l', 'Cp_u', 'Cp_k', 'Cpm'):
            print(f'{key:<5s} {getattr(self, key):8.4f}', end='')
            limits = getattr(self, f'{key}_limits')
            print(f'{limits[0]:8.4f}{limits[1]:8.4f}')
        print()
        print(f'Exp<LSL {self.exp_LSL:3.0f}%   Obs<LSL {self.obs_LSL:3.0f}%')
        print(f'Exp>USL {self.exp_USL:3.0f}%   Obs>USL {self.obs_USL:3.0f}%')

    def plot(self, bins='scott', ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))
        data = self.qcc.data
        if isinstance(data, pd.DataFrame):
            data = data.values
        data = data.flatten()
        data = data[~np.isnan(data)]
        xlim = [min(np.min(data), np.min(self.spec_limits.values), self.target),
                max(np.max(data), np.max(self.spec_limits.values), self.target)]
        xlim = xlim + np.diff(xlim) * [-0.1, 0.1]

        ax.hist(data, bins=bins, density=True, color='lightgrey', edgecolor='grey')
        xx = np.linspace(*xlim, num=250)
        ax.plot(xx, stats.norm(self.center, self.std_dev).pdf(xx), ls='--', color='gray')

        position = [self.spec_limits.LSL[0], self.target, self.spec_limits.USL[0]]
        ax.axvline(position[0], color='red', ls=':')
        ax.axvline(position[1], color='red', ls='--')
        ax.axvline(position[2], color='red', ls=':')

        ax.secondary_xaxis('bottom')
        ax.xaxis.tick_top()
        ax.set_xticks(position)
        ax.set_xticklabels(['LSL', 'Target', 'USL'])
        ax.set_yticks([])

        fig = ax.get_figure()
        fig.subplots_adjust(bottom=0.25)
        options = {'fontsize': 11}
        col = [0.1, 0.35, 0.55, 0.75]
        row = [0.18, 0.14, 0.1, 0.06, 0.02]
        fig.text(col[0], row[0], f'Number of obs = {self.nobs}', **options)
        fig.text(col[0], row[1], f'Center = {self.center:.5g}', **options)
        fig.text(col[0], row[2], f'StdDev = {self.std_dev:.5g}', **options)

        fig.text(col[1], row[0], f'Target = {self.center:.2g}', **options)
        fig.text(col[1], row[1], f'LSL = {self.spec_limits.LSL[0]:.2g}', **options)
        fig.text(col[1], row[2], f'USL = {self.spec_limits.USL[0]:.2g}', **options)

        fig.text(col[2], row[0], f'Cp = {self.Cp:.2g}', **options)
        fig.text(col[2], row[1], f'Cp_l = {self.Cp_l:.2g}', **options)
        fig.text(col[2], row[2], f'Cp_u = {self.Cp_u:.2g}', **options)
        fig.text(col[2], row[3], f'Cp_k = {self.Cp_k:.2g}', **options)
        fig.text(col[2], row[4], f'Cpm = {self.Cpm:.2g}', **options)

        fig.text(col[3], row[0], f'Exp<LSL = {self.exp_LSL:.2g}%', **options)
        fig.text(col[3], row[1], f'Exp>USL = {self.exp_USL:.2g}%', **options)
        fig.text(col[3], row[2], f'Obs<LSL = {self.obs_LSL:.2g}%', **options)
        fig.text(col[3], row[3], f'Obs>USL = {self.obs_USL:.2g}%', **options)
        return ax



from .bandit import optimalOAB, simulateOAB
from .dodge.dodge_chain import ChainPlanBinomial, ChainPlanPoisson
from .dodge.dodge_curtailed import curtailedBinomial
from .dodge.dodge_double import DSPlanBinomial, DSPlanPoisson
from .dodge.dodge_sequential import sequentialDesign
from .dodge.dodge_single import SSPlanBinomial, SSPlanHyper, SSPlanPoisson
from .dodge.dodge_other import lotSensitiveComplianceSampPlan, variableSampPlanKnown, variableSampPlanUnknown
from .generic import findPlan, findPlanApprox
from .oc import OperatingCharacteristics2c



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
from scipy.cluster.hierarchy import dendrogram, fcluster


def plot_dendrogram(model, **kwargs):
    # Create linkage matrix and then plot the dendrogram

    # create the counts of samples under each node
    counts = np.zeros(model.children_.shape[0])
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        current_count = 0
        for child_idx in merge:
            if child_idx < n_samples:
                current_count += 1  # leaf node
            else:
                current_count += counts[child_idx - n_samples]
        counts[i] = current_count

    linkage_matrix = np.column_stack([model.children_, model.distances_,
                                      counts]).astype(float)

    threshold = kwargs.get('color_threshold')
    if threshold is None and model.n_clusters_ is not None:
        # find threshold for given number of clusters
        threshold = np.max(model.distances_) / 2
        delta = threshold / 2
        while True:
            count = len(set(fcluster(linkage_matrix, t=threshold, criterion='distance')))
            if count == model.n_clusters_:
                break
            if count > model.n_clusters_:
                threshold = threshold + delta
            else:
                threshold = threshold - delta
            delta = delta / 2

    # Plot the corresponding dendrogram
    dendrogram(linkage_matrix, color_threshold=threshold, **kwargs)



# pylint: disable=line-too-long
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest

from mistat.simulation import PistonSimulator, simulationGroup
from mistat.simulation.mistatSimulation import repeat_elements


class TestPistonSimulation(unittest.TestCase):
    def test_repeat_elements(self):
        assert list(repeat_elements(1, 10)) == [1] * 10
        assert list(repeat_elements([1], 10)) == [1] * 10
        assert list(repeat_elements([1, 2], 10)) == [*([1] * 10), *([2] * 10)]
        assert list(repeat_elements([1, 2, 3], 10)) == [*([1] * 10), *([2] * 10), *([3] * 10)]

    def test_simulationGroup(self):
        simulator = PistonSimulator(seed=1)

        with pytest.raises(ValueError):
            simulationGroup(simulator, 10)

        simulation_result = simulator.simulate()
        assert simulation_result.shape == (50, 8)
        assert 'group' not in simulation_result

        for group_size, expected_length in (7, 49), (2, 50), (3, 48), (25, 50), (40, 40), (1, 50):
            grouped = simulationGroup(simulation_result, group_size, quiet=True)
            assert grouped.shape == (expected_length, 9)
            assert 'group' in grouped
            group_values = grouped['group'].value_counts()
            assert set(group_values.index) == set(range(1, expected_length // group_size + 1))
            assert all(count == group_size for count in group_values)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest

import mistat
from mistat import bayes


class TestBayes(unittest.TestCase):

    def test_updateBetaMixture(self):
        betaMixture = bayes.Mixture(
            probabilities=[0.5, 0.5],
            distributions=[
                bayes.BetaDistribution(a=1, b=1),
                bayes.BetaDistribution(a=15, b=2),
            ]
        )
        data = [10, 2]
        result = bayes.updateBetaMixture(betaMixture, data)
        assert result.probabilities[0] == pytest.approx(0.2845528)
        assert result.probabilities[1] == pytest.approx(0.7154472)
        assert result.distributions[0] == bayes.BetaDistribution(a=11, b=3)
        assert result.distributions[1] == bayes.BetaDistribution(a=25, b=4)
        assert result.distributions[0].theta() == pytest.approx(0.7857142)
        assert result.distributions[1].theta() == pytest.approx(0.86206896)

        betaMixture = bayes.Mixture(
            probabilities=[0.5, 0.25, 0.25],
            distributions=[
                bayes.BetaDistribution(a=1, b=1),
                bayes.BetaDistribution(a=15, b=2),
                bayes.BetaDistribution(a=2, b=22),
            ]
        )
        result = bayes.updateBetaMixture(betaMixture, data)
        assert result.probabilities[0] == pytest.approx(0.4430337)
        assert result.probabilities[1] == pytest.approx(0.5569567)
        assert result.probabilities[2] == pytest.approx(0.0000096, abs=1e-6)

    def test_updateGammaMixture(self):
        mixture = bayes.Mixture(
            probabilities=[0.5, 0.5],
            distributions=[
                bayes.GammaDistribution(shape=1, rate=1),
                bayes.GammaDistribution(shape=15, rate=2),
            ]
        )
        data = {'y': [5], 't': [1]}
        result = bayes.updateGammaMixture(mixture, data)
        assert result.distributions[0] == bayes.GammaDistribution(shape=6, rate=2)
        assert result.distributions[1] == bayes.GammaDistribution(shape=20, rate=3)
        assert result.probabilities[0] == pytest.approx(0.1250978)
        assert result.probabilities[1] == pytest.approx(0.8749022)

        data = {'y': [5, 3], 't': [1, 2]}
        result = bayes.updateGammaMixture(mixture, data)
        assert result.distributions[0] == bayes.GammaDistribution(shape=9, rate=4)
        assert result.distributions[1] == bayes.GammaDistribution(shape=23, rate=5)
        assert result.probabilities[0] == pytest.approx(0.8127316)
        assert result.probabilities[1] == pytest.approx(0.1872684)

        mixture = bayes.Mixture(
            probabilities=[0.5, 0.25, 0.25],
            distributions=[
                bayes.GammaDistribution(shape=1, rate=1),
                bayes.GammaDistribution(shape=15, rate=2),
                bayes.GammaDistribution(shape=10, rate=3),
            ]
        )
        result = bayes.updateGammaMixture(mixture, data)
        assert result.probabilities[0] == pytest.approx(0.33694729)
        assert result.probabilities[1] == pytest.approx(0.03881946)
        assert result.probabilities[2] == pytest.approx(0.62423325)



# mypy: disallow_untyped_defs,disallow_untyped_calls
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import string
from pathlib import Path
from typing import Dict, Union

import pandas as pd

DATA_DIR = Path(__file__).parent / 'csvFiles'

DatasetType = Union[pd.DataFrame, pd.Series, Dict[str, pd.Series]]


def load_data(name: str) -> DatasetType:
    """ Returns the data either as a Pandas data frame or series """
    data_file = get_data_file(name)
    if not data_file.exists():
        raise ValueError('Data file {name} not found')
    if name in SPECIAL_DATASETS:
        return SPECIAL_DATASETS[name](data_file)
    data = pd.read_csv(data_file)
    if data.shape[1] == 1:
        return data[data.columns[0]]  # pylint: disable=unsubscriptable-object
    return data


def describe_data(name: str) -> str:
    """ Return information about the data file """
    description_file = get_description_file(name)
    if not description_file.exists():
        raise ValueError('Description for data file {name} not found')
    text = description_file.read_text()
    text = text.replace('R Documentation', 'Documentation')
    return text


def get_data_file(name: str) -> Path:
    if name.endswith('.gz'):
        name = name[:-3]
    if name.endswith('.csv'):
        name = name[:-4]
    return DATA_DIR / f'{name}.csv.gz'


def get_description_file(name: str) -> Path:
    if name.endswith('.gz'):
        name = name[:-3]
    if name.endswith('.csv'):
        name = name[:-4]
    if name.endswith('.md'):
        name = name[:-3]
    description_file = DATA_DIR / 'md' / f'{name}.md'
    if not description_file.exists():
        name = description_file.with_suffix('').name.rstrip(string.digits)
        description_file = DATA_DIR / 'md' / f'{name}.md'
    return description_file


def load_process_segment(data_file: Path) -> Dict[str, pd.Series]:
    data = pd.read_csv(data_file)
    return {column: data[column].dropna() for column in data.columns}


SPECIAL_DATASETS = {
    'PROCESS_SEGMENT': load_process_segment,
}



# pylint: disable=too-many-arguments,too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
from scipy import stats

from .dodge_base import AcceptanceSamplingPlan


def DSPlanBinomial(N, n1, n2, Ac1, Re1, Ac2, p=None):
    p = p if p is not None else np.arange(0, 0.255, 0.005)
    p = np.array(p)

    Pa1 = stats.binom.cdf(Ac1, n1, p)
    limits = list(range(Ac1 + 1, Re1))

    Pa2 = np.zeros([len(p), len(limits)])
    for i, d1 in enumerate(limits):
        Pa2[:, i] = stats.binom.pmf(d1, n1, p) * stats.binom.cdf(Ac2-d1, n2, p)
    Pa2 = np.sum(Pa2, axis=1)
    OC = Pa1 + Pa2

    ASN = n1 + n2 * (stats.binom.cdf((Re1 - 1), n1, p) -
                     stats.binom.cdf(Ac1, n1, p))
    AOQ = (p * Pa1 * (N - n1) + p * Pa2 * (N - n1 - n2)) / N
    ATI = n1 * Pa1 + (n1 + n2) * Pa2 + (1 - OC) * N

    return AcceptanceSamplingPlan(p=p, OC=OC, AOQ=AOQ, ATI=ATI, ASN=ASN, Pa1=Pa1, Pa2=Pa2)


def DSPlanPoisson(N, n1, n2, Ac1, Re1, Ac2, p=None):
    p = p if p is not None else np.arange(0, 0.255, 0.005)
    p = np.array(p)

    Pa1 = stats.poisson.cdf(Ac1, n1*p)
    limits = list(range(Ac1 + 1, Re1))

    Pa2 = np.zeros(len(p))
    for d1 in limits:
        Pa2 += stats.poisson.pmf(d1, n1*p) * stats.poisson.cdf(Ac2-d1, n2*p)

    OC = Pa1 + Pa2
    ASN = n1 + n2 * (stats.poisson.cdf((Re1 - 1), n1*p) -
                     stats.poisson.cdf(Ac1, n1*p))
    AOQ = (p * Pa1 * (N - n1) + p * Pa2 * (N - n1 - n2)) / N
    ATI = n1 * Pa1 + (n1 + n2) * Pa2 + (1 - OC) * N
    return AcceptanceSamplingPlan(p=p, OC=OC, AOQ=AOQ, ATI=ATI, ASN=ASN, Pa1=Pa1, Pa2=Pa2)


def DSPlanNormal(N, n1, n2, Ac1, Re1, Ac2, p=None):
    p = p if p is not None else np.arange(0, 0.255, 0.005)
    p = np.array(p)

    dn1 = np.sqrt(n1*p*(1-p)*(1 - n1/N))
    dn2 = np.sqrt(n2*p*(1-p)*(1 - n2/(N*n1)))

    Pa1 = stats.norm.cdf((Ac1 + 0.5 - n1*p) / dn1)

    limits = list(range(Ac1 + 1, Re1))

    Pa2 = np.zeros(len(p))
    for d1 in limits:
        delta = (stats.norm.cdf((d1 + 0.5 - n1*p) / dn1) -
                 stats.norm.cdf((d1 - 0.5 - n1*p) / dn1))
        Pa2 += delta * stats.norm.cdf((Ac2-d1 + 0.5 - n2*p)/dn2)

    OC = Pa1 + Pa2
    ASN = n1 + n2 * (stats.norm.cdf((Re1 - 0.5 - n1*p)/dn1) -
                     stats.norm.cdf((Ac1 + 0.5 - n1*p)/dn1))
    AOQ = (p * Pa1 * (N - n1) + p * Pa2 * (N - n1 - n2)) / N
    ATI = n1 * Pa1 + (n1 + n2) * Pa2 + (1 - OC) * N

    return AcceptanceSamplingPlan(p=p, OC=OC, AOQ=AOQ, ATI=ATI, ASN=ASN, Pa1=Pa1, Pa2=Pa2)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np

from .dodge_base import SequentialSamplePlan


def sequentialDesign(AQL, alpha, LQL, beta, oc_type='binomial', N=None):
    """ Create a sequential sampling plan

    Selects the appropriate sequential sampling plan from the given inputs
    """
    # R: dodge.SeqDesignBinomial
    a = np.log((1 - beta) / alpha)
    b = np.log((1 - alpha) / beta)
    g1 = np.log(LQL / AQL)
    g2 = np.log((1 - AQL) / (1 - LQL))
    G = g1 + g2
    h1 = b / G
    h2 = a / G
    s = g2 / G

    h = np.arange(-4 * h1, 5 * h2, 0.01)

    f1 = (1 - LQL) / (1 - AQL)
    f2 = LQL / AQL
    p = (1 - f1 ** h) / ((f2 ** h) - (f1 ** h))
    L = int(round(2 * h1 / s))
    k = list(range(1, L + 1))
    accept = s * np.array(k) - h1
    reject = s * np.array(k) + h2

    return SequentialSamplePlan(AQL=AQL, alpha=alpha, LQL=LQL, beta=beta, N=N, oc_type=oc_type, h1=h1, h2=h2, s=s,
                                h=h, k=k, accept=accept, reject=reject, p=p)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors  # type: ignore
from matplotlib.patches import Rectangle


class ParetoChart:
    def __init__(self, data, title=None, labels=None):
        self.title = title
        if labels is None:
            self.labels = data.index
        else:
            self.labels = labels
        self.data = pd.DataFrame({
            'Labels': self.labels,
            'Frequency': np.array(data).flatten()
        })
        self.data = self.data.sort_values(by='Frequency', ascending=False)
        self.data['Cum.Freq.'] = np.cumsum(self.data['Frequency'])
        total = np.sum(self.data['Frequency'])
        self.data['Percentage'] = 100 * self.data['Frequency'] / total
        self.data['Cum.Percent.'] = 100 * self.data['Cum.Freq.'] / total

    def print(self):
        if self.title is None:
            print('Pareto chart analysis')
        else:
            print(f'Pareto chart analysis: {self.title}')
        print(self.data)

    def plot(self, ax=None, rotation=None, ha='center'):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))

        nlevels = len(self.data)
        cumfreq = self.data['Cum.Freq.']
        cumperc = [0, 25, 50, 75, 100]
        q = [perc / 100 * max(cumfreq) for perc in cumperc]

        ax.plot(range(0, nlevels), self.data['Cum.Freq.'],
                marker='o', markerfacecolor='black', color='grey')
        ax.bar(range(0, nlevels), self.data['Frequency'])
        greys = plt.get_cmap('gray')(np.linspace(0, 1, nlevels + 3))[2:-1]
        bars = [child for child in ax.get_children() if isinstance(child, Rectangle)]
        for barObj, color in zip(bars, greys):
            barObj.set_color(colors.to_hex(color))
            barObj.set_edgecolor('black')

        ax.set_xticks(range(0, nlevels))
        ax.set_xticklabels(self.data['Labels'], rotation=rotation, ha=ha)
        xlim = ax.get_xlim()
        ax.hlines(q, *ax.get_xlim(), linestyle='dotted', color='grey')
        ax.set_xlim(*xlim)

        ax.set_ylim(0, max(cumfreq) * 1.05)
        secax = ax.secondary_yaxis('right')
        secax.yaxis.tick_right()
        secax.set_ylabel('Frequency')
        ax.set_ylabel('Cumulative Percentage')
        ax.set_yticks(q)
        ax.set_yticklabels(f'{p} %' for p in cumperc)

        ax.get_figure().suptitle(self.title or 'Pareto chart analysis', fontsize=14)
        return ax



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import math
from collections import namedtuple

import numpy as np
from scipy import stats
from scipy.optimize import root_scalar

from mistat.acceptanceSampling.distributions import OChypergeom
from mistat.acceptanceSampling.oc import OC_TYPES

RiskPoint = namedtuple('RiskPoint', 'pdefect,paccept')
Plan = namedtuple('Plan', 'n,c,r')
PlanNormal = namedtuple('PlanNormal', 'n,k,s_type')


def findPlan(PRP, CRP, oc_type='binomial', N=None, s_type='known'):
    """ Find the sampling plan with the smallest sample size, which meets a prespecified
    Producer and Consumer Risk Points.

    The convention used here, as in many books, is to use equality for the Producer Risk
    Point rather than the consumer risk point.

    No consideration is given to "cost functions
    """
    oc_type = oc_type.lower()
    if oc_type not in [*OC_TYPES, 'normal']:
        raise ValueError(f'Unknown type {oc_type}')
    PRP = checkRiskPoint(PRP, 'Producer risk point', oc_type)
    CRP = checkRiskPoint(CRP, 'Consumer risk point', oc_type)
    if np.any(PRP[0] >= CRP[0]):
        raise ValueError('Consumer Risk Point quality must be greater than Producer Risk Point quality')

    if oc_type == 'binomial':
        raise NotImplementedError
        # c <- 0
        # n <- c+1
        # repeat {
        #   if (calc.OCbinomial(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
        #     n <- n + 1
        #   else if (calc.OCbinomial(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
        #     c <- c + 1
        #   else
        #     break
        # }
        # return(list(n=n, c=c, r=c+1))

    if oc_type == 'hypergeom':
        if N is None:
            raise ValueError('N must be supplied for the hypergeometric distribution.')
        c = np.array([0])
        n = c + 1
        while True:
            if OChypergeom(n=n, c=c, r=c + 1, N=N, D=CRP[0] * N).paccept > CRP[1]:
                n += 1
            elif OChypergeom(n=n, c=c, r=c + 1, N=N, D=PRP[0] * N).paccept < PRP[1]:
                c += 1
            else:
                break
        return Plan(n[0], c[0], (c + 1)[0])

    if oc_type == 'poisson':
        raise NotImplementedError
        # c = 0
        # n = c + 1
        # while True:
        #     if OCpoisson(n=[n], c=[c], r=[c + 1], pd=CRP[0]).paccept > CRP[1]:
        #         print(n, c, OCpoisson(n=[n], c=[c], r=[c + 1], pd=CRP[0]).paccept)
        #         n += 1
        #     elif OCpoisson(n=[n], c=[c], r=[c + 1], pd=PRP[0]).paccept < PRP[1]:
        #         print(n, c, OCpoisson(n=[n], c=[c], r=[c + 1], pd=PRP[0]).paccept)
        #         c += 1
        #     else:
        #         break
        #     break
        # return Plan(n, c, c + 1)

    if oc_type == 'normal':
        s_type = s_type.lower()
        if s_type not in ['known', 'unknown']:
            raise ValueError(f'Unknown s_type {s_type}')
        # With known standard deviation
        if s_type == 'known':
            n = stats.norm.ppf(1 - PRP[1])[0] + stats.norm.ppf(CRP[1])[0]
            n = n / (stats.norm.ppf(CRP[0])[0] - stats.norm.ppf(PRP[0])[0])
            n = int(np.ceil(n * n))
            k = stats.norm.ppf(1 - PRP[1])[0] / np.sqrt(n) - stats.norm.ppf(PRP[0])[0]
            return PlanNormal(n, k, s_type)
        if s_type == 'unknown':
            n = 2
            k = find_k(n, PRP[0], PRP[1], interval=[0, 1000])
            nc = - stats.norm.ppf(CRP[0])[0] * np.sqrt(n)
            pa = 1 - stats.nct.cdf(k * np.sqrt(n), df=n - 1, nc=nc)
            while pa > CRP[1]:
                n = n + 1
                k = find_k(n, PRP[0], PRP[1], interval=[0, 1000])
                nc = - stats.norm.ppf(CRP[0])[0] * np.sqrt(n)
                pa = 1 - stats.nct.cdf(k * np.sqrt(n), df=n - 1, nc=nc)
            return PlanNormal(n, k, s_type)
    raise NotImplementedError


def findPlanApprox(PRP, CRP, N):
    """ Calculate single-stage sampling plan using approximation of hypergeometric distribution """
    alpha = (1 - PRP[1])
    beta = CRP[1]
    p0 = PRP[0]
    pt = CRP[0]

    za_pq0 = stats.norm.ppf(1 - alpha) * np.sqrt(p0 * (1 - p0))
    zb_pqt = stats.norm.ppf(1 - beta) * np.sqrt(pt * (1 - pt))

    # calculate approximation of n
    n0 = ((za_pq0 + zb_pqt) / (pt - p0)) ** 2
    nstar = int(round(n0 / (1 + n0 / N)))

    # calculate approximation of c
    cstar = int(round(nstar * p0 - 0.5 + za_pq0 * np.sqrt(nstar * (1 - nstar / N))))
    return Plan(nstar, cstar, cstar + 1)


def checkRiskPoint(RP, info, oc_type):
    if not isinstance(RP, list) and len(RP) != 2:
        raise ValueError(f'{info} must be a list with two elements (pdefect, paccept)')
    pdefect, paccept = RP
    if not isinstance(pdefect, list):
        pdefect = [pdefect]
    if not isinstance(paccept, list):
        paccept = [paccept]
    pdefect = np.array(pdefect)
    paccept = np.array(paccept)
    if oc_type == 'poisson':
        if not np.all(0 < pdefect):
            raise ValueError(f'{info}: all quality values of must be larger than 0')
        if not np.all(0 < paccept):
            raise ValueError(f'{info}: all paccept values of must be larger than 0')
    else:
        if not np.all(0 < pdefect < 1):
            raise ValueError(f'{info}: all values of pdefect must fall within [0, 1]')
        if not np.all(0 < paccept < 1):
            raise ValueError(f'{info}: all values of paccept must fall within [0, 1]')

    return RiskPoint(pdefect, paccept)


def find_k(n, pd, pa, interval=None):
    """ find k for a given n """
    if interval is None:
        interval = [0, 5]
    df = n - 1
    sqrt_n = np.sqrt(n)
    nc = - stats.norm.ppf(pd) * sqrt_n
    pa = 1 - pa

    def f(x):
        return stats.nct.cdf(x * sqrt_n, df=df, nc=nc) - pa
    return root_scalar(f, bracket=interval, method='brentq').root


# """
# find.plan <- function(PRP, CRP,
#                       type=c("binomial","hypergeom","poisson","normal"),
#                       N,
#                       s.type=c("known", "unknown"))
# {

#   else if(CRP[1] <= PRP[1])
#     stop("Consumer Risk Point quality must be greater than Producer Risk Point quality")

#   ## Attributes Sampling Plan - Binomial distribution
#   if (type == "binomial") {
#     c <- 0
#     n <- c+1
#     repeat {
#       if (calc.OCbinomial(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
#         n <- n + 1
#       else if (calc.OCbinomial(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
#         c <- c + 1
#       else
#         break
#     }
#     return(list(n=n, c=c, r=c+1))
#   }
#   ## Attributes Sampling Plan - Hypergeometric distribution
#   if (type == "hypergeom") {
#     c <- 0
#     n <- c+1
#     repeat {
#       if (calc.OChypergeom(n=n,c=c,r=c+1,N=N,D=CRP[1]*N) > CRP[2])
#         n <- n + 1
#       else if (calc.OChypergeom(n=n,c=c,r=c+1,N=N,D=PRP[1]*N) < PRP[2])
#         c <- c + 1
#       else
#         break
#     }
#     return(list(n=n, c=c, r=c+1))
#   }
#   ## Attributes Sampling Plan - Poisson distribution
#   if (type == "poisson") {
#     c <- 0
#     n <- c+1
#     repeat {
#       if (calc.OCpoisson(n=n,c=c,r=c+1,pd=CRP[1]) > CRP[2])
#         n <- n + 1
#       else if (calc.OCpoisson(n=n,c=c,r=c+1,pd=PRP[1]) < PRP[2])
#         c <- c + 1
#       else
#         break
#     }
#     return(list(n=n, c=c, r=c+1))
#   }
#   ## Variables Sampling Plan - Normal distribution
#   else if (type=="normal") {
#     ## With known standard deviation
#     if (s.type=="known") {
#       n <- ceiling( ((qnorm(1-PRP[2]) + qnorm(CRP[2]))/
#                      (qnorm(CRP[1])-qnorm(PRP[1])) )^2)
#       k <- qnorm(1-PRP[2])/sqrt(n) - qnorm(PRP[1])
#       return(list(n=n, k=k, s.type=s.type))
#     }
#     ## With unknown standard deviation
#     else if (s.type=="unknown") {
#       n <- 2 ## Need a minimum of 1 degree of freedom (=n-1) for the NC t-dist
#       k <- find.k(n, PRP[1], PRP[2], interval=c(0,1000))
#       pa <- 1- pt(k*sqrt(n), df=n-1, ncp=-qnorm(CRP[1])*sqrt(n))
#       while(pa > CRP[2]){
#         n <- n+1
#         k <- find.k(n, PRP[1], PRP[2])
#         pa <- 1-pt(k*sqrt(n), df=n-1, ncp=-qnorm(CRP[1])*sqrt(n))
#       }
#       return(list(n=n, k=k, s.type=s.type))
#     }
#   }
# }
# """


def normalApproximationOfH(j, M, n, N):
    P = N / M
    Q = 1 - P
    return stats.norm.cdf((j + 0.5 - n * P) / math.sqrt(n * P * Q * (1 - n / M)))



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass, fields
from typing import Optional

import numpy as np
import pandas as pd
from scipy import stats

from mistat.simulation.mistatSimulation import SimulationResult

from .mistatSimulation import (MistatSimulation, convert_to_list,
                               repeat_elements)

errors = {
    'm': 0.1,
    's': 0.01,
    'v0': 0.00025,
    'k': 50,
    'p0': 0.01,
    't': 0.13,
    't0': 0.13
}


@dataclass
class PowerCircuitSimulation(MistatSimulation):  # pylint: disable=too-many-instance-attributes
    rsA: float = 8_200
    rsB: float = 220_000
    rsC: float = 1_000
    rsD: float = 33_000
    rsE: float = 56_000
    rsF: float = 5_600
    rsG: float = 3_300
    rsH: float = 58.5
    rsI: float = 1_000
    rsJ: float = 120
    trK: float = 130
    trL: float = 100
    trM: float = 130
    tlA: float = 5
    tlB: float = 10
    tlC: float = 10
    tlD: float = 5
    tlE: float = 5
    tlF: float = 5
    tlG: float = 10
    tlH: float = 5
    tlI: float = 5
    tlJ: float = 5
    tlK: float = 5
    tlL: float = 10
    tlM: float = 5

    n_simulation: int = 50  # desired number of simulations
    seed: Optional[float] = None
    check: bool = True

    def __post_init__(self):
        if self.seed is not None:
            np.random.seed(seed=self.seed)

        self.resistors = [field.name for field in fields(self)
                          if field.name.startswith('rs')]
        self.transistors = [field.name for field in fields(self)
                            if field.name.startswith('tr')]
        self.tolerances = [field.name for field in fields(self)
                           if field.name.startswith('tl')]
        self.options = [*self.resistors, *self.transistors, *self.tolerances]

        for option in self.options:
            validate_range(getattr(self, option), 0, f'{option} must be greater than 0')

        if self.n_simulation < 1:
            raise ValueError('Number of simulations must be greater 1')

        # Convert to lists
        maxsize = 0
        for option in self.options:
            values = convert_to_list(getattr(self, option))
            maxsize = max(maxsize, len(values))
            setattr(self, option, values)
        if maxsize == 1:
            maxsize = self.n_simulation

        # Make sure that the vectors are all the same length
        for option in self.options:
            values = getattr(self, option)
            if maxsize % len(values) != 0:
                raise ValueError(f'Inconsistent length of option {option}')
            values = np.array(list(repeat_elements(values, maxsize // len(values))))
            setattr(self, option, values)

        self.df = pd.DataFrame({option: getattr(self, option) for option in self.options})

    def simulate(self):
        size = len(self.rsA)

        # add errors
        data = {}
        for option in self.options:
            if option.startswith('tl'):
                element = f'rs{option[2]}' if hasattr(self, f'rs{option[2]}') else f'tr{option[2]}'
                rv = stats.norm.rvs(size=size)
                rv = rv * 2 * getattr(self, option) * getattr(self, element) / 600
                data[element] = data[element] + rv
            else:
                data[option] = getattr(self, option)

        # calculate resulting voltage
        data['volts'] = self.vCircuit(**data)

        # copy specified values into results
        for option in self.resistors:
            data[option] = getattr(self, option)
        for option in self.transistors:
            data[option] = getattr(self, option)
        return SimulationResult(data)

    @staticmethod
    def vCircuit(rsA, rsB, rsC, rsD, rsE, rsF, rsG, rsH, rsI, rsJ, trK, trL, trM):  # pylint: disable=too-many-arguments
        a = rsB / (rsA + rsB)
        b = (rsA * rsB / (rsA + rsB) + rsC) / (trL * trM) + rsI
        c = rsE + 0.5 * rsG
        d = rsA * rsB * trK / (rsA + rsB)
        e = rsF + 0.5 * rsG
        f = (c + e) * (1 + trK) * rsH + c * e
        g = rsH + 0.6
        h = 1.2

        aden = 1 + d * e / f + b * (1 / rsJ + 0.006 * (1 + 13.67 / rsJ)) + a * 0.1367 * 0.6
        anum = (a + b / rsJ) * (138 - 1.33) + d * (c + e) * g / f - h

        return anum / aden


def validate_range(value, left, message):
    left = left - 1e-6
    if isinstance(value, (tuple, list, pd.core.series.Series, np.ndarray)):
        if not all(left <= v for v in value):
            raise ValueError(message)
    else:
        if not left <= value:
            raise ValueError(message)



# pylint: disable=line-too-long
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest

from mistat.simulation.pistonSimulation import (PistonSimulator,
                                                uniformSumDistribution)


class TestPistonSimulation(unittest.TestCase):
    def test_uniformSumDistribution(self):
        for size in range(1, 10):
            rvs = uniformSumDistribution(size=size)
            assert len(rvs) == size
            assert all(0 <= rv <= 1 for rv in rvs)

        for size in range(1, 10):
            rvs = uniformSumDistribution(size=size, left=-3, right=0)
            assert len(rvs) == size
            assert all(-3 <= rv <= 0 for rv in rvs)

    def test_cycleTime(self):
        assert PistonSimulator.cycleTime(60, 0.02, 0.01, 5000, 110000, 296, 360) == pytest.approx(0.3859461)
        assert PistonSimulator.cycleTime(60, 0.02, 0.01, 5000, 110000, 100000, 360) == pytest.approx(0.4813737)
        assert PistonSimulator.cycleTime(60, 0.02, 0.01, 5000, 110000, 296, 3600) == pytest.approx(0.2260668)
        assert PistonSimulator.cycleTime(70, 0.02, 0.01, 5000, 110000, 296, 360) == pytest.approx(0.4168694)
        assert PistonSimulator.cycleTime(60, 0.2, 0.01, 5000, 110000, 296, 360) == pytest.approx(0.06090253)
        assert PistonSimulator.cycleTime(60, 0.2, 0.01, 5000, 110000, 296, 360) == pytest.approx(0.06090253)

    def test_PistonSimulator(self):
        s = [0.005] * 100
        s.extend([0.01] * 100)
        s.extend([0.02] * 100)

        simulator = PistonSimulator(seed=1, s=s)

        result = simulator.simulate()
        return result
#         plt.scatter(x=result.index, y=result.seconds)
#         plt.show()
#
#         np.random.seed(0)
#         t = [296] * 35
#         t.extend([296 * np.power(1.1, k) for k in range(1, 66)])
#         simulator = PistonSimulator(t=t, check=False, seed=1236)
#         result = simulator.simulate()
#         print(result)
#
#         plt.scatter(x=result.index, y=result.seconds)
#         plt.show()



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
__version__ = '0.1.6'



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import math


def stemLeafDiagram(X, rowsPerGroup, leafUnit=1.0, latex=False):
    sep = '&' if latex else ''

    X = X.sort_values(ignore_index=True)
    size = len(X)
    cumCount = 0
    revCumCount = size
    lastGroupNr = None
    for _, group in X.groupby([int(xi * rowsPerGroup / (leafUnit * 10)) for xi in X]):
        count = len(group)
        cumCount += count
        # round down
        group = group.apply(lambda xi: math.floor(xi / leafUnit))
        digits = ''.join(f'{xi}'[-1] for xi in group)

        cumFreq = min(cumCount, revCumCount)
        groupNr = int(group.iloc[0]/10)
        if lastGroupNr and lastGroupNr < groupNr - 1:
            # add lines with no values
            while lastGroupNr < groupNr - 1:
                lastGroupNr += 1
                s = '0'
                print(f'{s:>8s} {sep} {lastGroupNr:-5d} {sep}')
        lastGroupNr = groupNr
        if (cumCount >= size // 2) and (revCumCount >= size // 2):
            s = f'({count})'
            print(f'{s:>8s} {sep} {groupNr:-5d} {sep}  {digits}')
        else:
            print(f'{cumFreq:-8d} {sep} {groupNr:-5d} {sep}  {digits}')
        if latex:
            print('\\\\')
        revCumCount -= count



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import random

import numpy as np
import pandas as pd
import pingouin as pg


def availabilityEBD(ttf, ttr, n, seed=None, printSummary=True):
    ''' given TTF and TTR, determines the availbility EBD and calculates asymptotic availability '''
    if seed is not None:
        np.random.seed(seed)
    ttf = np.array(ttf)
    ttr = np.array(ttr)

    _, asf = pg.compute_bootci(ttf, func='mean', n_boot=1000, return_dist=True)
    _, asr = pg.compute_bootci(ttr, func='mean', n_boot=1000, return_dist=True)
    bootavail = pd.Series(asf / (asf + asr), name='availability EBD')
    if printSummary:
        avf = np.mean(asf)
        avr = np.mean(asr)
        avail = avf / (avf + avr)
        print(f'The estimated MTTF from ttf is {avf:.2f}')
        print(f'The estimated MTTR from ttr is {avr:.2f}')
        print(f'The estimated asymptotic availability is {avail:.4f}')
        print()
        print(bootavail.describe())
    return bootavail


def renewalEBD(ttf, ttr, time, n, seed=None, printSummary=True):
    ''' given TTF and TTR, determines the EBD of the number of renewals '''
    if seed is not None:
        random.seed(seed)
    ttf = np.array(ttf)
    ttr = np.array(ttr)

    result = []
    for _ in range(n):
        tt = 0
        it = 1
        while tt < time:
            tt += random.choice(ttf)
            tt += random.choice(ttr)
            it += 1
        result.append(it)

    result = pd.Series(result, name='number of renewals EBD')
    if printSummary:
        print(f'The estimated MEAN NUMBER Of RENEWALS is {np.mean(result):.2f}')
        print(result.describe())
    return result




def addTreatments(design, mainEffects):
    """ Add treatments information to design matrix

    design is a pandas dataframe with the design matrix as created by the doepy package
    mainEffects is a list of the columns that define main effects
    """
    design = design.copy()
    mainEffects = list(mainEffects)
    # collect information about high values for each main effect
    high = {effect: design[effect].max() for effect in mainEffects}
    treatments = []
    for _, row in design.iterrows():
        treatment = ''.join([effect for effect in mainEffects if row[effect] == high[effect]])
        if not treatment:
            treatment = '(1)'
        treatments.append(treatment)
    design.insert(0, 'Treatments', treatments)
    return design


def _reduceTreatment(treatment):
    """ Reduce a treatment string that can contain same effects multiple times """
    effects = []
    for c in sorted(set(treatment)):
        if treatment.count(c) % 2 == 1:
            effects.append(c)
    return ''.join(sorted(effects))


def subgroupOfDefining(defining, noTreatment=''):
    """ Given the defining treatments of a design, enumerate the subgroup """
    subgroup = set(defining)
    for d in defining:
        subgroup.update([_reduceTreatment(d + sg) for sg in subgroup])
    subgroup.remove('')
    subgroup.add(noTreatment)
    return sorted(subgroup)


def aliasesInSubgroup(effect, defining):
    subgroup = subgroupOfDefining(defining)
    aliases = set()
    for treatment in subgroup:
        if not treatment:
            continue
        aliases.add(_reduceTreatment(effect + treatment))
    return sorted(aliases)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest

import mistat
from mistat import runsTest


class TestRunsTest(unittest.TestCase):

    def test_runsTest(self):
        x = mistat.load_data('RNORM10')
        x = [0 if xi < 10 else 1 for xi in x]

        assert runsTest(x, alternative='less').statistic == pytest.approx(-0.35956, rel=1e-5)
        assert runsTest(x, alternative='less').pval == pytest.approx(0.3596, rel=1e-4)
        assert runsTest(x).pval == pytest.approx(0.7192, rel=1e-4)
        assert runsTest(x, alternative='greater').pval == pytest.approx(0.6404, rel=1e-4)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import pytest

from mistat.acceptanceSampling.bandit import (optimalOAB, optimalOAB2,
                                              simulateOAB)


class TestBandit(unittest.TestCase):

    def test_simulateOAB(self):
        result = simulateOAB(50, 0.4, 0.5, 10, 0.95, 100, seed=1)
        assert result.reward.mean == 22.215
        assert result.mgamma.mean == 31.57

    def test_optimalOAB(self):
        result = optimalOAB(10, 0.5)
        assert result.rewards[-1, 0] == pytest.approx(5.823719)
        assert result.max_reward == pytest.approx(0.5 * (5.823719 + 10))

        result = optimalOAB(50, 0.5)
        assert result.max_reward == pytest.approx(40.17491)

        result = optimalOAB2(10, 0.5)
        assert result.rewards[-1, 0] == pytest.approx(5.823719)
        assert result.max_reward == pytest.approx(0.5 * (5.823719 + 10))

        result = optimalOAB2(50, 0.5)
        assert result.max_reward == pytest.approx(40.17491)



# pylint: disable=line-too-long
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest
from pathlib import Path

import numpy as np
import pandas as pd

from mistat.qcc.qualityControlChart import QualityControlChart, qcc_groups
from mistat.qcc.statistics import QCCStatistics, SD_estimator
from mistat.qcc.tests.utilities import assertQCC, assertQCCviolations


class TestComparisonWithR(unittest.TestCase):
    def test_pistonrings(self):
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]
        # newdata = diameter[25:, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])
        assertQCCviolations(qcc)
        # qcc(diameter[1:25,], type="xbar", newdata=newdata)
        # assertNewQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])
        qcc = QualityControlChart(data, qcc_type='xbar', nsigmas=2)
        assertQCC(qcc, 74.00118, 0.009785039, [73.99242, 74.00993])
        assertQCCviolations(qcc, beyondLimits={'UCL': [0], 'LCL': [13]})
        # qcc(diameter[1:25,], type="xbar", newdata=newdata)
        # assertNewQCC(qcc, 74.00118, 0.009785039, [73.98805, 74.0143])

        qcc = QualityControlChart(data, qcc_type='R')
        assertQCC(qcc, 0.02276, 0.009785039, [0, 0.04812533])

        qcc = QualityControlChart(data, qcc_type='S')
        assertQCC(qcc, 0.009240037, 0.009829977, [0, 0.019302416])

    def test_pistonrings_variable_control_limits(self):
        """
        # variable control limits
        out <- c(9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100)
        diameter <- qcc.groups(pistonrings$diameter[-out], sample[-out])
        r <- qcc(diameter[1:25,], type="xbar")
        r$center
        r$std.dev
        r$limits
        r <- qcc(diameter[1:25,], type="R")
        qcc(diameter[1:25,], type="S")
        qcc(diameter[1:25,], type="xbar", newdata=diameter[26:40,])
        qcc(diameter[1:25,], type="R", newdata=diameter[26:40,])
        qcc(diameter[1:25,], type="S", newdata=diameter[26:40,])
        """
        out = [idx - 1 for idx in [9, 10, 30, 35, 45, 64, 65, 74, 75, 85, 99, 100]]
        pistonrings = pd.read_csv(Path(__file__).parent / 'data' / 'pistonrings.csv')
        pistonrings = pistonrings.drop(pistonrings.index[out])
        diameter = qcc_groups(pistonrings['diameter'], pistonrings['sample'])
        data = diameter[:25, ]
        # newdata = diameter[25:, ]

        qcc = QualityControlChart(data, qcc_type='xbar')
        assertQCC(qcc, 74.00075, 0.009856701, [[73.98753, 74.01398], [73.98368, 74.01782], [73.98753, 74.01398]])
        assertQCCviolations(qcc)
        qcc = QualityControlChart(data, qcc_type='xbar', nsigmas=2)
        assertQCC(qcc, 74.00075, 0.009856701, [[73.99194, 74.00957], [73.98937, 74.01213], [73.99194, 74.00957]])
        assertQCCviolations(qcc, beyondLimits={'UCL': [0], 'LCL': [13]})

        qcc = QualityControlChart(data, qcc_type='R')
        assertQCC(qcc, 0.02230088, 0.009856701, [[0, 0.04785198], [0, 0.04857007], [0, 0.04785198]])

        qcc = QualityControlChart(data, qcc_type='S')
        assertQCC(qcc, 0.00938731, 0.009930923, [[0, 0.01955302], [0, 0.02318885], [0, 0.01955302]])

    def test_orangejuice(self):
        """
        data(orangejuice)
        attach(orangejuice)
        r <- qcc(D[trial], sizes=size[trial], type="p")

        # remove out-of-control points (see help(orangejuice) for the reasons)
        inc <- setdiff(which(trial), c(15,23))
        q1 <- qcc(D[inc], sizes=size[inc], type="p")

        r <- qcc(D[inc], sizes=size[inc], type="p", newdata=D[!trial], newsizes=size[!trial])
        r$center
        r$std.dev
        r$limits
        detach(orangejuice)
        """
        oj = pd.read_csv(Path(__file__).parent / 'data' / 'orangejuice.csv')
        data = oj[oj['trial']]
        # newdata = oj[~oj['trial']]

        qcc = QualityControlChart(data['D'], sizes=data['size'], qcc_type='p')
        assertQCC(qcc, 0.2313333, 0.421685, [0.05242755, 0.4102391])
        assertQCCviolations(qcc, beyondLimits={'UCL': [14, 22], 'LCL': []})

        dataRed = data.drop(data.index[qcc.violations['beyondLimits']['UCL']])
        qcc = QualityControlChart(dataRed['D'], sizes=dataRed['size'], qcc_type='p')
        assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])
        assertQCCviolations(qcc, beyondLimits={'UCL': [19], 'LCL': []})

        # qcc = QualityControlChart(dataRed['D'], newdata=newdata, sizes=dataRed['size'], qcc_type='p')
        # assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])

    def test_orangejuice2(self):
        """
        data(orangejuice2)
        attach(orangejuice2)
        names(D) <- sample
        r = qcc(D[trial], sizes=size[trial], type="p")
        q2 <- qcc(D[trial], sizes=size[trial], type="p", newdata=D[!trial], newsizes=size[!trial])
        detach(orangejuice2)
        """
        oj = pd.read_csv(Path(__file__).parent / 'data' / 'orangejuice2.csv')
        data = oj[oj['trial']]
        # newdata = oj[~oj['trial']]

        qcc = QualityControlChart(data['D'], sizes=data['size'], qcc_type='p')
        assertQCC(qcc, 0.1108333, 0.3139256, [0, 0.2440207])
        assertQCCviolations(qcc)

        # qcc = QualityControlChart(dataRed['D'], newdata=newdata, sizes=dataRed['size'], qcc_type='p')
        # assertQCC(qcc, 0.215, 0.4108223, [0.04070284, 0.3892972])

    def test_xbar_one(self):
        # viscosity data (Montgomery, pag. 242)
        x = [33.75, 33.05, 34, 33.81, 33.46, 34.02, 33.68, 33.27, 33.49, 33.20, 33.62, 33.00, 33.54, 33.12, 33.84]
        qcc = QualityControlChart(x, qcc_type="xbarone")
        assertQCC(qcc, 33.52333, 0.4261651, [32.24484, 34.80183])

        qcc = QualityControlChart(x, qcc_type="xbarone", std_dev=SD_estimator.sd)
        assertQCC(qcc, 33.52333, 0.3415928, [32.49855, 34.54811])

        #  Water content of antifreeze data (Wetherill and Brown, 1991, p. 120)
        x = [2.23, 2.53, 2.62, 2.63, 2.58, 2.44, 2.49, 2.34, 2.95, 2.54, 2.60, 2.45, 2.17, 2.58, 2.57, 2.44, 2.38,
             2.23, 2.23, 2.54, 2.66, 2.84, 2.81, 2.39, 2.56, 2.70, 3.00, 2.81, 2.77, 2.89, 2.54, 2.98, 2.35, 2.53]
        qcc = QualityControlChart(x, qcc_type="xbarone")
        assertQCC(qcc, 2.569706, 0.1794541, [2.031344, 3.108068])
        assertQCCviolations(qcc)

        qcc = QualityControlChart(x, qcc_type="xbarone", std_dev='SD')
        assertQCC(qcc, 2.569706, 0.2216795, [1.904667, 3.234744])
        assertQCCviolations(qcc)

        expected = [0.1794541, 0.1897519, 0.1955224, 0.2019203, 0.2022154, 0.2041948, 0.2072357, 0.2121212, 0.2126056, 0.2143082, 0.2172579,
                    0.2189612, 0.2183180, 0.2178859, 0.2181558, 0.2172365, 0.2168714, 0.2151667, 0.2138331, 0.2128866, 0.2122988, 0.2108173,
                    0.2095927]
        XbarOne = QCCStatistics().get('xbarone')
        np.testing.assert_allclose([XbarOne.sd(x, k=k) for k in range(2, 25)], expected, rtol=1e-5)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest
from unittest.case import skip

import numpy as np
import pandas as pd

from mistat.acceptanceSampling.generic import (Plan, PlanNormal, find_k,
                                               findPlan, findPlanApprox)


class TestGeneric(unittest.TestCase):
    def test_findPlan(self):
        result = findPlan(PRP=(0.01, 0.05), CRP=(0.08, 0.05), oc_type="hypergeom", N=100)
        assert result == Plan(31, 0, 1)

        p0 = np.array([*[0.01] * 10, *[0.03] * 10])
        pt = np.linspace(0.05, 0.32, 10)
        pt = np.array([*pt, *pt])

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.05), CRP=(pti, 0.05), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [45, 31, 23, 18, 15, 13, 11, 10, 9, 8] * 2)
        np.testing.assert_array_equal(result['c'], [0] * 20)

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.1), CRP=(pti, 0.2), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [27, 18, 13, 11, 9, 7, 6, 6, 5, 5] * 2)
        np.testing.assert_array_equal(result['c'], [0] * 20)

        result = []
        for p0i, pti in zip(p0, pt):
            result.append([p0i, pti, *findPlan(PRP=(p0i, 0.5), CRP=(pti, 0.3), oc_type="hypergeom", N=100)])
        result = pd.DataFrame(result, columns=['p0', 'pt', 'n', 'c', 'r'])
        np.testing.assert_array_equal(result['n'], [21, 14, 10, 8, 7, 6, 5, 4, 4, 4, 42, 14, 10, 8, 7, 6, 5, 4, 4, 4])
        np.testing.assert_array_equal(result['c'], [*[0] * 10, 1, *[0] * 9])

    def test_findPlan_normal(self):
        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="normal", N=500)
        assert result == PlanNormal(55, 2.1045558562282713, 'known')

        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="normal", N=500, s_type='unknown')
        assert result == PlanNormal(176, 2.105480330251979, 'unknown')

    @skip('Not implemented')
    def test_findPlan_poisson(self):
        result = findPlan(PRP=(0.01, 0.95), CRP=(0.03, 0.05), oc_type="poisson", N=500)
        assert result == Plan(524, 9, 10)

    def test_findPlanApprox(self):
        result = findPlanApprox(PRP=(0.01, 0.95), CRP=(0.03, 0.05),  N=500)
        assert result == Plan(248, 4, 5)
        result = findPlanApprox(PRP=(0.01, 0.95), CRP=(0.05, 0.05),  N=500)
        assert result == Plan(127, 2, 3)

    def test_find_k(self):
        assert find_k(100, 0.01, 0.95, [0, 1000]) == 2.0400552555161156
        assert find_k(10, 0.01, 0.95, [0, 1000]) == 1.562533782230915



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
import pandas as pd
from scipy.linalg import eig
from scipy.optimize import NonlinearConstraint, minimize


class ResponseSurfaceMethod:
    ''' Class provides information about the response surface defined by a
    second order regression model

     See section on "Canonical representation" in book for mathematical details'''

    def __init__(self, model, codes):
        ''' determine vector b and array B from model '''
        b = np.array([model.params.get(c, 0) for c in codes])
        B = np.zeros([len(codes), len(codes)])
        for i, code_i in enumerate(codes):
            B[i, i] = model.params.get(f'np.power({code_i}, 2)', 0)
            for j, code_j in enumerate(codes[i + 1:], i + 1):
                B[i, j] = model.params.get(f'{code_i}:{code_j}', 0) / 2
                B[j, i] = B[i, j]
        self.model = model
        self.codes = codes
        self.b = b
        self.B = B

    def f(self, x):
        ''' calculate function value at position x '''
        exog = pd.Series(x, index=self.codes)
        return self.model.predict(exog=exog)[0]

    def gradient(self, x):
        ''' calculate gradient at position x '''
        gradient = pd.Series(self.b + np.matmul(self.B, x), index=self.codes)
        return gradient / np.linalg.norm(gradient)

    def stationary_point(self):
        return pd.Series(- 0.5 * np.matmul(np.linalg.inv(self.B), self.b), index=self.codes)

    def _eigen(self):
        ev, evec = eig(self.B)
        return {'eval': ev, 'evec': evec}

    def eigen(self):
        ev, evec = eig(self.B)
        indx = ev.real.argsort()[::-1]
        return {'eval': ev.real[indx], 'evec': evec[:, indx]}

    def constrainedOptimization(self, start, distances=None, maximize=True, reverse=False):
        if distances is None:
            distances = np.arange(0.5, 5.1, 0.5)
        x0 = start
        if x0 is None:
            x0 = self.stationary_point()
        else:
            x0 = pd.Series(x0, index=self.codes)

        factor = -1 if maximize else 1

        def fun(x):
            df = pd.DataFrame({code: [value] for code, value in zip(self.codes, x)})
            return factor * self.model.predict(df)
        result = {'step': 0, 'distance': 0}
        result.update({c: x0[c] for c in self.codes})
        result['y'] = factor * fun(x0)[0]
        results = [result]
        x = x0
        lastDistance = 0
        for step, distance in enumerate(distances, 1):
            result = {'step': 1, 'distance': distance}
            gradient = self.gradient(x)
            if reverse:
                gradient = -gradient
                reverse = False
            x = x + (distance-lastDistance) * gradient

            def distanceConstraint(x):
                return np.sqrt(np.sum((x-x0)**2)) - distance  # pylint: disable=cell-var-from-loop

            constraints = NonlinearConstraint(distanceConstraint, -0.0001, 0.0001)
            minResult = minimize(fun, x, constraints=constraints)
            x = pd.Series(minResult.x, index=self.codes)
            result = {'step': step, 'distance': distance}
            result.update({c: x[c] for c in self.codes})
            result['y'] = factor * minResult.fun
            results.append(result)
        return pd.DataFrame(results)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from mistat.acceptanceSampling.distributions import OCtype, getDistribution

OC_TYPES = ['binomial', 'hypergeom', 'poisson']


# TODO valdate default handling
class OperatingCharacteristics2c:
    def __init__(self, sample_sizes, acceptance_numbers, rejection_numbers=None,
                 oc_type='binomial', **kwargs):
        if oc_type.lower() not in OC_TYPES:
            raise ValueError(f'Unknown type {oc_type}')
        self.oc_type = oc_type.lower()
        if isinstance(sample_sizes, int):
            sample_sizes = [sample_sizes]
        if isinstance(acceptance_numbers, int):
            acceptance_numbers = [acceptance_numbers]

        self.samples = sample_sizes
        self.acceptance = acceptance_numbers
        self.rejection = rejection_numbers

        nacceptance = len(self.acceptance)
        if self.rejection is None and nacceptance <= 2:
            self.rejection = [1 + self.acceptance[-1]] * nacceptance
        elif isinstance(self.rejection, int):
            self.rejection = [self.rejection]

        self.nstages = len(self.samples)
        if any([len(self.rejection) != self.nstages, len(self.acceptance) != self.nstages]):
            raise ValueError('Inconsistent length of arguments sample_sizes, acceptance_numbers and rejection_numbers')

        if self.acceptance[-1] + 1 != self.rejection[-1]:
            raise ValueError('Decision from last sample cannot be made: r != c+1')

        if self.oc_type == 'binomial':
            self.distribution = getDistribution(self.samples, self.acceptance,
                                                OCtype.binomial, r=self.rejection, **kwargs)
        elif self.oc_type == 'hypergeom':
            self.distribution = getDistribution(self.samples, self.acceptance,
                                                OCtype.hypergeom, r=self.rejection, **kwargs)

    @property
    def pd(self):
        return self.distribution.pd

    @property
    def paccept(self):
        return self.distribution.paccept



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import matplotlib.pyplot as plt
import numpy as np
import pingouin
from pingouin.plotting import _ppoints
from scipy import stats


def qqplot_patched(x, dist='norm', sparams=(), confidence=.95, figsize=(5, 4),
                   ax=None):
    """Quantile-Quantile plot.
    Parameters
    ----------
    x : array_like
        Sample data.
    dist : str or stats.distributions instance, optional
        Distribution or distribution function name. The default is `'norm'`
        for a normal probability plot.
    sparams : tuple, optional
        Distribution-specific shape parameters (shape parameters, location,
        and scale). See :py:func:`scipy.stats.probplot` for more details.
    confidence : float
        Confidence level (.95 = 95%) for point-wise confidence envelope.
        Can be disabled by passing False.
    figsize : tuple
        Figsize in inches
    ax : matplotlib axes
        Axis on which to draw the plot
    Returns
    -------
    ax : Matplotlib Axes instance
        Returns the Axes object with the plot for further tweaking.
    Raises
    ------
    ValueError
        If ``sparams`` does not contain the required parameters for ``dist``.
        (e.g. :py:class:`scipy.stats.t` has a mandatory degrees of
        freedom parameter *df*.)
    Notes
    -----
    This function returns a scatter plot of the quantile of the sample data
    ``x`` against the theoretical quantiles of the distribution given in
    ``dist`` (default = *'norm'*).
    The points plotted in a Q–Q plot are always non-decreasing when viewed
    from left to right. If the two distributions being compared are identical,
    the Q–Q plot follows the 45° line y = x. If the two distributions agree
    after linearly transforming the values in one of the distributions,
    then the Q–Q plot follows some line, but not necessarily the line y = x.
    If the general trend of the Q–Q plot is flatter than the line y = x,
    the distribution plotted on the horizontal axis is more dispersed than
    the distribution plotted on the vertical axis. Conversely, if the general
    trend of the Q–Q plot is steeper than the line y = x, the distribution
    plotted on the vertical axis is more dispersed than the distribution
    plotted on the horizontal axis. Q–Q plots are often arced, or "S" shaped,
    indicating that one of the distributions is more skewed than the other,
    or that one of the distributions has heavier tails than the other.
    In addition, the function also plots a best-fit line (linear regression)
    for the data and annotates the plot with the coefficient of
    determination :math:`R^2`. Note that the intercept and slope of the
    linear regression between the quantiles gives a measure of the relative
    location and relative scale of the samples.
    .. warning:: Be extra careful when using fancier distributions with several
        parameters. Always double-check your results with another
        software or package.
    References
    ----------
    * https://github.com/cran/car/blob/master/R/qqPlot.R
    * Fox, J. (2008), Applied Regression Analysis and Generalized Linear
      Models, 2nd Ed., Sage Publications, Inc.
    Examples
    --------
    Q-Q plot using a normal theoretical distribution:
    .. plot::
        >>> import numpy as np
        >>> import pingouin as pg
        >>> np.random.seed(123)
        >>> x = np.random.normal(size=50)
        >>> ax = pg.qqplot(x, dist='norm')
    Two Q-Q plots using two separate axes:
    .. plot::
        >>> import numpy as np
        >>> import pingouin as pg
        >>> import matplotlib.pyplot as plt
        >>> np.random.seed(123)
        >>> x = np.random.normal(size=50)
        >>> x_exp = np.random.exponential(size=50)
        >>> fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(9, 4))
        >>> ax1 = pg.qqplot(x, dist='norm', ax=ax1, confidence=False)
        >>> ax2 = pg.qqplot(x_exp, dist='expon', ax=ax2)
    Using custom location / scale parameters as well as another Seaborn style
    .. plot::
        >>> import numpy as np
        >>> import seaborn as sns
        >>> import pingouin as pg
        >>> import matplotlib.pyplot as plt
        >>> np.random.seed(123)
        >>> x = np.random.normal(size=50)
        >>> mean, std = 0, 0.8
        >>> sns.set_style('darkgrid')
        >>> ax = pg.qqplot(x, dist='norm', sparams=(mean, std))
    """
    if isinstance(dist, str):
        dist = getattr(stats, dist)

    x = np.asarray(x)
    x = x[~np.isnan(x)]  # NaN are automatically removed

    # Check sparams: if single parameter, tuple becomes int
    if not isinstance(sparams, (tuple, list)):
        sparams = (sparams,)
    # For fancier distributions, check that the required parameters are passed
    if len(sparams) < dist.numargs:
        raise ValueError("The following sparams are required for this " +
                         f"distribution: {dist.shapes}. See scipy.stats.{dist.name} for details.")

    # Extract quantiles and regression
    quantiles = stats.probplot(x, sparams=sparams, dist=dist, fit=False)
    theor, observed = quantiles[0], quantiles[1]

    fit_params = dist.fit(x)
    loc = fit_params[-2]
    scale = fit_params[-1]
    shape = fit_params[:-2] if len(fit_params) > 2 else None

    # Observed values to observed quantiles
    if loc != 0 and scale != 1:
        observed = (np.sort(observed) - fit_params[-2]) / fit_params[-1]

    # Linear regression
    slope, intercept, r, _, _ = stats.linregress(theor, observed)

    # Start the plot
    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=figsize)

    ax.plot(theor, observed, 'bo')

    ax.set_xlabel('Theoretical quantiles')
    ax.set_ylabel('Ordered quantiles')
    ax.set_title('Q-Q Plot')

    # Add diagonal line
    end_pts = [ax.get_xlim(), ax.get_ylim()]
    end_pts[0] = min(end_pts[0])
    end_pts[1] = max(end_pts[1])
    ax.plot(end_pts, end_pts, color='slategrey', lw=1.5)
    ax.set_xlim(end_pts)
    ax.set_ylim(end_pts)

    # Add regression line and annotate R2
    fit_val = slope * theor + intercept
    ax.plot(theor, fit_val, 'r-', lw=2)
    posx = end_pts[0] + 0.60 * (end_pts[1] - end_pts[0])
    posy = end_pts[0] + 0.10 * (end_pts[1] - end_pts[0])
    ax.text(posx, posy, f'$R^2={r**2:.3f}$')

    if confidence is not False:
        # Confidence envelope
        n = x.size
        P = _ppoints(n)
        crit = stats.norm.ppf(1 - (1 - confidence) / 2)
        pdf = dist.pdf(theor) if shape is None else dist.pdf(theor, *shape)
        se = (slope / pdf) * np.sqrt(P * (1 - P) / n)
        upper = fit_val + crit * se
        lower = fit_val - crit * se
        ax.plot(theor, upper, 'r--', lw=1.25)
        ax.plot(theor, lower, 'r--', lw=1.25)

    return ax


pingouin.qqplot = qqplot_patched



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import statsmodels.formula.api as smf

from mistat.rsm.responseSurfaceMethod import ResponseSurfaceMethod


class TestResponseSurfaceMethod(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data1 = loadModel('simulationData1.csv')
        cls.data2 = loadModel('simulationData2.csv')

    def test_function(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        assert rsm.f([0, 0, 0, 0]) == pytest.approx(0.559, abs=0.001)
        assert rsm.f([1.260, -1.244, 0.870, 0.318]) == pytest.approx(0.415, abs=0.001)

        model = self.data2['model']
        codes = self.data2['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        assert rsm.f([0, 0, 0, 0]) == pytest.approx(0.586, abs=0.001)
        assert rsm.f([1.808, -0.167, 0.449, -0.705]) == pytest.approx(0.406, abs=0.001)

    def test_stationary(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        np.testing.assert_allclose(rsm.stationary_point(), np.array(
            [-10.642838, 2.957658, 3.660820, -2.299632]), rtol=1e-4)

        model = self.data2['model']
        codes = self.data2['codes']
        rsm = ResponseSurfaceMethod(model, codes)
        np.testing.assert_allclose(rsm.stationary_point(), np.array(
            [-1.1969668, 0.9320400, 2.2297196, 0.4233004]), rtol=1e-4)

    def test_eigen(self):
        model = self.data1['model']
        codes = self.data1['codes']
        rsm = ResponseSurfaceMethod(model, codes)

        # first test the complex eigenvalues
        eigen = rsm._eigen()  # pylint: disable=protected-access
        ev = sorted(eigen['eval'].real, reverse=True)
        np.testing.assert_allclose(ev, np.array(
            [0.034424442, 0.008661927, 0.003316576, -0.001943580]), rtol=1e-4)
        ev = sorted(ev.imag for ev in eigen['eval'])
        np.testing.assert_allclose(ev, np.zeros(4), rtol=1e-4)

        # nowtest the sorted real eigenvalues/vectors
        eigen = rsm.eigen()
        np.testing.assert_allclose(eigen['eval'], np.array(
            [0.034424442, 0.008661927, 0.003316576, -0.001943580]), rtol=1e-4)
        evec1 = eigen['evec'][:, 0]
        if evec1[0] < 0:
            evec1 = -evec1
        np.testing.assert_allclose(evec1, np.array(
            [0.2722108, 0.2597354, 0.9261707, 0.0254271]), rtol=1e-4)


def loadModel(filename):
    factors = {
        's': [0.01, 0.015],
        'v0': [0.00625, 0.00875],
        'k': [2000, 4000],
        't0': [345, 355],
    }
    result = pd.read_csv(Path(__file__).parent / filename)
    # calculate mean and std of response by group
    result = result.groupby(by='group')
    result = result.agg({'s': 'mean', 'v0': 'mean', 'k': 'mean',
                         't0': 'mean', 'seconds': ['mean', 'std']})
    result.columns = ['s', 'v0', 'k', 't0', 'Ymean', 'Ystd']

    # transformation between factors and code levels
    factor2x = {factor: f'x{i}' for i, factor in enumerate(factors, 1)}
    x2factor = {f'x{i}': factor for i, factor in enumerate(factors, 1)}
    center = {factor: 0.5 * (max(values) + min(values)) for factor, values in factors.items()}
    unit = {factor: 0.5 * (max(values) - min(values)) for factor, values in factors.items()}

    # # define helper function to convert code co-ordinates to factor co-ordinates
    # def toFactor(code, codeValue):
    #     ''' convert code to factor co-ordinates '''
    #     factor = x2factor[code]
    #     return center[factor] + codeValue * unit[factor]

    # add code levels to table
    for c in factors:
        result[factor2x[c]] = (result[c] - center[c]) / unit[c]
    formula = ('Ymean ~ (x1+x2+x3+x4)**2 + ' +
               'np.power(x1,2) + np.power(x2,2) + np.power(x3,2) + np.power(x4,2)')
    return {'codes': list(x2factor), 'model': smf.ols(formula, data=result).fit()}


class TestRSMpackage(unittest.TestCase):
    def test_RidgeRisingSituation(self):
        data = pd.DataFrame({
            'x1': [-1, 1, -1, 1, -1, 1, 0, 0, 0, 0, 0],
            'x2': [-1, -1, 1, 1, 0, 0, -1, 1, 0, 0, 0],
            'Response': [52.3, 5.3, 46.7, 44.2, 58.5, 33.5, 32.8, 49.2, 49.3, 50.2, 51.6],
        })
        codes = ['x1', 'x2']
        model = smf.ols('Response ~ (x1 + x2)**2 + np.power(x1,2) + np.power(x2,2)', data=data).fit()
        rsm = ResponseSurfaceMethod(model, codes)
        np.testing.assert_allclose(rsm.stationary_point(),
                                   np.array([-5.176505, -2.706733]), rtol=1e-4)
        eigen = rsm.eigen()
        np.testing.assert_allclose(eigen['eval'],
                                   np.array([-0.509419, -12.706370]), rtol=1e-4)
        np.testing.assert_allclose(eigen['evec'][:, 0],
                                   np.array([0.8396245, 0.5431673]), rtol=1e-4)
        np.testing.assert_allclose(eigen['evec'][:, 1],
                                   np.array([-0.5431673, 0.8396245]), rtol=1e-4)

        path = rsm.constrainedOptimization(start=(0, 0), distances=(0.5, 1))
        np.testing.assert_allclose(path['y'],
                                   np.array([50.263158, 55.556944, 58.742308]), rtol=1e-4)



'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import numpy as np
import pandas as pd


def plotECDF(data, ax=None):
    n = len(data)
    ecdf = pd.DataFrame({
        'x': np.sort(data),
        'Fn(x)': range(1, n + 1),
    })
    ecdf['Fn(x)'] = ecdf['Fn(x)'] / n
    ax = ecdf.plot(x='x', y='Fn(x)', color='black', drawstyle='steps-post', legend=False, ax=ax)
    ax.axhline(y=0, color='grey', linestyle='--')
    ax.axhline(y=1, color='grey', linestyle='--')
    ax.set_ylabel('Fn(x)')
    return ax



# pylint: disable=line-too-long
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import unittest

import numpy as np
import pytest

from mistat.qcc.shro import (runLengthShroNorm, runLengthShroPois,
                             shroArlPfaCedNorm, shroArlPfaCedPois)


class TestCusum(unittest.TestCase):
    def test_shroArlPfaCedNorm(self):
        x = [-1.0080707,  1.3549394, -0.4689749,  1.4681936,  0.4425564, 0.1462031,  0.1715031,
             0.5925072,  2.7647493, 0.6192188, 0.3945324, -1.6215868, -0.9459741, -0.9978776]

        res = runLengthShroNorm(x, 1, 1, 10, 1, 19)
        assert res['rl'] == 9
        assert len(res['w']) == 9
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [2.344282e-01, 3.469896e-09, 7.275562e-01, 4.415837e-05, 1.319939e-06, 1.699855e-06, 1.145011e-04, 3.110184e+05]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 1.001, 1, 0.58)
        assert res['rl'] == 7
        assert len(res['w']) == 7
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.3683780, 0.5031622, 0.5537946, 0.5718620, 0.5783402, 0.5807385]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 1.001, 0.9, 0.58)
        assert res['rl'] == 4
        assert len(res['w']) == 4
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.4071206, 0.5718245, 0.6399946]):
            assert e1 == pytest.approx(e2)

        res = runLengthShroPois(x, 11 / 6, 1, 19)
        assert res['rl'] == np.inf
        assert len(res['w']) == 14
        assert res['w'][0] is None
        for e1, e2 in zip(res['w'][1:], [0.8363378, 0.5083987, 1.3511733, 1.1310696, 0.8566233, 0.7578374, 0.9260974, 3.7859976, 2.5626072,  1.6646799, 0.3668443, 0.2834033, 0.25786117291]):
            assert e1 == pytest.approx(e2)

        res = shroArlPfaCedNorm(seed=1, N=10, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(135.0)
        assert res['Std. Error'] == pytest.approx(40.54873)

        res = shroArlPfaCedNorm(seed=1, N=10, mean0=10, tau=10, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(10.7)
        assert res['Std. Error'] == pytest.approx(0.51088159)
        assert res['PFA'] == pytest.approx(0.1)
        assert res['CED'] == pytest.approx(1.22222222)
        assert res['CED-Std. Error'] == pytest.approx(3.72107039)

    def test_shroArlPfaCedPois(self):
        x = [12, 8, 15, 13, 11, 10, 8, 10, 17, 16,
             89, 85, 120, 98, 109, 110, 93, 85, 98, 109]
        res = runLengthShroPois(x, 1.1, 1, 19)
        assert res['rl'] == 10

        x = [7, 7, 6, 22, 16, 11, 11, 16, 9, 12,
             112, 121, 95, 98, 114, 103, 99, 97, 86, 105]
        res = runLengthShroPois(x, 1.1, 1, 19)
        assert res['rl'] == 8

        res = shroArlPfaCedPois(seed=1, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(24.79)
        assert res['Std. Error'] == pytest.approx(1.570178)

        res = shroArlPfaCedPois(seed=1, tau=10, lambda1=100, verbose=False)['statistic']
        assert res['ARL'] == pytest.approx(10.89)
        assert res['Std. Error'] == pytest.approx(0.044486, abs=0.0001)
        assert res['PFA'] == pytest.approx(0.03)
        assert res['CED'] == pytest.approx(0.958763)
        assert res['CED-Std. Error'] == pytest.approx(1.108611)



# pylint: disable=too-many-arguments,too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
from dataclasses import dataclass
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np


@dataclass
class AcceptanceSamplingPlan:
    p: List[float]
    OC: List[float]
    AOQ: List[float]
    ASN: Optional[List[float]] = None
    ATI: Optional[List[float]] = None
    n: Optional[List[float]] = None
    Pa1: Optional[List[float]] = None
    Pa2: Optional[List[float]] = None

    def plot(self, axes=None):
        if axes is None:
            _, axes = plt.subplots(nrows=2, ncols=2, figsize=[8, 8])

        samplePlan = self

        ax = axes[0, 0]
        ax.plot(self.p, samplePlan.OC)
        ax.set_xlabel('Fraction Nonconforming p')
        ax.set_ylabel('Probability of Acceptance')
        ax.set_title('OC Curve')

        ax = axes[0, 1]
        if self.ASN is not None:
            ax.plot(self.p, samplePlan.ASN)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('Average sample size')
            ax.set_title(f'maximum ASN = {max(samplePlan.ASN):.0f}')

        if self.n is not None:
            ax.plot(self.p, self.n)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('Uncurtailed sample size')

        ax = axes[1, 0]
        ax.plot(self.p, samplePlan.AOQ)
        ax.set_xlabel('Fraction Nonconforming p')
        ax.set_ylabel('AOQ')
        ax.set_title(f'AOQL = {max(samplePlan.AOQ):.4g}')

        if self.ATI is not None:
            ax = axes[1, 1]
            ax.plot(self.p, self.ATI)
            ax.set_xlabel('Fraction Nonconforming p')
            ax.set_ylabel('ATI')
        else:
            axes[1, 1].set_axis_off()


@dataclass
class SequentialSamplePlan:
    # R: dodge.SeqDesignBinomial
    AQL: float  # Acceptable quality level
    alpha: float  # alpha producer’s risk
    LQL: float  # LQL Limiting quality level
    beta: float  # beta consumers’ risk

    h1: float
    h2: float
    s: float
    h: List[float]
    p: List[float]

    k: List[int]
    accept: List[float]
    reject: List[float]

    N: Optional[int] = None  # lot size, required for hypergeometric
    oc_type: str = 'binomial'

    def __post_init__(self):
        k1 = ((1 - self.beta) / self.alpha) ** self.h
        k2 = (self.beta / (1 - self.alpha)) ** self.h
        self.OC = (k1 - 1) / (k1 - k2)
        self.AOQ = self.p * self.OC

        k5 = self.OC * np.log(self.beta / (1 - self.alpha))
        k3 = k5 + (1 - self.OC) * np.log((1 - self.beta) / self.alpha)
        k4 = self.p * np.log(self.LQL / self.AQL) + (1 - self.p) * np.log((1 - self.LQL) / (1 - self.AQL))
        self.ASN = k3 / k4

        self.ATI = self.N and k5 / k4 + (1 - self.OC) * self.N

    def plot(self):
        asp = AcceptanceSamplingPlan(p=self.p, OC=self.OC, ASN=self.ASN, AOQ=self.AOQ, ATI=self.ATI, n=None)
        asp.plot()

    def acceptanceChart(self, ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=[5, 5])
        ax.plot(self.k, self.accept, color='C0')
        ax.plot(self.k, self.reject, color='C0')
        ax.grid(linestyle='--', color='lightgrey')

        ax.set_ylim(min(self.accept), max(self.reject))
        ax.set_xlabel('k')
        ax.set_ylabel(r'$d_k$')
        ax.set_title('Sequential Acceptance Chart')

        x = np.median(self.k)
        common = {'horizontalalignment': 'center', 'verticalalignment': 'center'}
        ax.text(x, 0.5 * (np.min(self.accept) + np.mean(self.accept)), "ACCEPT", **common)
        ax.text(x, 0.5 * (np.max(self.reject) + np.mean(self.reject)), "REJECT", **common)
        ax.text(x, max(self.accept), "CONTINUE", **common)



# pylint: disable=too-many-arguments,too-many-instance-attributes
'''
Modern Statistics: A Computer Based Approach with Python
Industrial Statistics: A Computer Based Approach with Python

(c) 2022 Ron Kenett, Shelemyahu Zacks, Peter Gedeck
'''
import matplotlib.pyplot as plt
import pandas as pd
from scipy import linalg

from mistat.mqcc.statistics import mqccStatistics
from mistat.qcc.rules import shewhartRules


class MultivariateQualityControlChart:
    def __init__(self, data, qcc_type=mqccStatistics.default,
                 center=None, cov=None,
                 limits=True, pred_limits=False, rules=None,
                 data_name=None, labels=None,
                 newdata=None, newlabels=None,
                 confidence_level=None):
        self.statistic = mqccStatistics.get(qcc_type)
        self.qcc_type = self.statistic.qcc_type

        self.sizes = self.statistic.get_sizes(data)
        num_samples, sample_sizes, p = self.sizes
        if isinstance(num_samples, list):
            raise ValueError('varying number of samples (rows)')
        if isinstance(sample_sizes, list):
            raise ValueError('varying sample size (columns)')
        if sample_sizes == 1:
            self.statistic = mqccStatistics.get('T2single')
            self.qcc_type = self.statistic.qcc_type

        if confidence_level is None:
            confidence_level = (1 - 0.0027) ** p
        if not 0 < confidence_level < 1:
            raise ValueError('confidence.level must be a numeric value in the range (0,1)')

        if labels is None:
            labels = list(range(len(data)))
        self.labels = labels

        self.stats = self.statistic.stats(data, center=center, cov=cov)

        if newdata is not None:
            self.newdata = newdata
            self.newsizes = self.statistic.get_sizes(newdata)
            self.newstats = self.statistic.stats(newdata, center=self.stats.center, cov=self.stats.cov)
            self.newlabels = newlabels
            if newlabels is None:
                self.newlabels = list(range(len(data), len(data) + len(newdata)))
            # gives self.newdata, self.newstats, self.newsizes
            pred_limits = True
        else:
            self.newdata = None
            self.newstats = None
            self.newsizes = None

        if isinstance(limits, bool):
            if limits:
                limits = self.statistic.limits(num_samples, sample_sizes, p, confidence_level)
                self.limits = limits['control']
            else:
                self.limits = None
        else:
            self.limits = limits

        if isinstance(pred_limits, bool):
            if pred_limits:
                pred_limits = self.statistic.limits(num_samples, sample_sizes, p, confidence_level)
                self.pred_limits = pred_limits['prediction']
            else:
                self.pred_limits = None
        else:
            self.pred_limits = pred_limits

        self.violations = shewhartRules(self, run_length=0)
        self.violations['beyondPredLimits'] = shewhartRules(
            self, run_length=0, limits=self.limits)['beyondLimits']

    def plot(self, title=None, ax=None, show_legend=True):
        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))
        beyondLimits = [*self.violations['beyondLimits']['LCL'], *self.violations['beyondLimits']['UCL']]
        violatingRuns = self.violations['violatingRuns']
        df = pd.DataFrame({'x': self.labels, 'y': self.stats.statistics})
        ax = df.plot.line(x='x', y='y', style='-', color='lightgrey',
                          marker='o', markerfacecolor='black', ax=ax)
        if self.newdata is not None:
            new_df = pd.DataFrame({'x': self.newlabels, 'y': self.newstats.statistics})
            ax = new_df.plot.line(x='x', y='y', style='-', color='lightgrey',
                                  marker='o', markerfacecolor='black', ax=ax)
            df = pd.concat([df, new_df])
            ax.axvline(0.5 * (self.newlabels[0] + self.labels[-1]), color='black', linestyle='--')

        ax.plot(df.iloc[beyondLimits]['x'], df.iloc[beyondLimits]['y'], linestyle='None',
                marker='s', markerfacecolor='red', markeredgecolor='red')
        ax.plot(df.iloc[violatingRuns]['x'], df.iloc[violatingRuns]['y'],
                linestyle='None', marker='s', markerfacecolor='red', markeredgecolor='red')

        ax.legend().remove()
        ax.set_xlabel('Group')
        ax.set_ylabel('Group summary statistics')

        yticks = {'ticks': [], 'labels': []}
        if self.limits is not None and len(self.limits) == 1:
            ax.axhline(self.limits.LCL[0], color='black', linestyle='--')
            ax.axhline(self.limits.UCL[0], color='black', linestyle='--')

            yticks['ticks'].extend([self.limits.LCL[0], self.limits.UCL[0]])
            yticks['labels'].extend(['LCL', 'UCL'])

        if self.pred_limits is not None and len(self.pred_limits) == 1:
            ax.axhline(self.pred_limits.LPL[0], color='black', linestyle=':')
            ax.axhline(self.pred_limits.UPL[0], color='black', linestyle=':')

            if yticks['ticks'] and yticks['ticks'][0] == self.pred_limits.LPL[0]:
                yticks['labels'][0] = 'LCL/LPL'
            else:
                yticks['labels'].append('LPL')
                yticks['ticks'].append(self.pred_limits.LPL[0])

            yticks['labels'].append('UPL')
            yticks['ticks'].append(self.pred_limits.UPL[0])

        if yticks['ticks']:
            ax2 = ax.twinx()
            ax2.set_ylim(*ax.get_ylim())
            ax2.yaxis.tick_right()
            ax2.set_yticks(yticks['ticks'])
            ax2.set_yticklabels(yticks['labels'])

        fig = ax.get_figure()
        if show_legend:
            fig.subplots_adjust(bottom=0.2)
            fig.text(0.1, 0.1, f'Number of groups = {len(self.labels)}', fontsize=12)
            fig.text(0.1, 0.06, f'Sample size = {self.sizes.samples_sizes}', fontsize=12)
            fig.text(0.1, 0.02, f'|S| = {linalg.det(self.stats.cov):.5g}', fontsize=12)
            if self.limits is not None and len(self.limits) == 1:
                fig.text(0.4, 0.1, f'LCL = {self.limits.LCL[0]:.5g}', fontsize=12)
                fig.text(0.4, 0.06, f'UCL = {self.limits.UCL[0]:.5g}', fontsize=12)
            if self.pred_limits is not None and len(self.pred_limits) == 1:
                fig.text(0.6, 0.1, f'LPL = {self.pred_limits.LPL[0]:.5g}', fontsize=12)
                fig.text(0.6, 0.06, f'UPL = {self.pred_limits.UPL[0]:.5g}', fontsize=12)
            fig.text(0.4, 0.02, f'Number beyond limits = {len(beyondLimits)}', fontsize=12)

        if self.newdata is not None:
            ymax = max(*self.stats.statistics, *self.newstats.statistics)
            ymin = min(*self.stats.statistics, *self.newstats.statistics)
            ymax = 1.05 * ymax - 0.05 * ymin
            xcalibration = 0.5 * (self.labels[0] + self.labels[-1])
            xnewdata = 0.5 * (self.newlabels[0] + self.newlabels[-1])
            ax.text(xcalibration, ymax, 'Calibration data', ha='center', va='top')
            ax.text(xnewdata, ymax, 'New data', ha='center', va='top')

        fig_title = [f'{self.statistic.qcc_type} Chart']
        if title is not None:
            fig_title.append(title)
        fig.suptitle('\n'.join(fig_title), fontsize=14)
        return ax
